"""# Waveforms example

Simple waveform generator widget, with plotting.
"""

from dataclasses import dataclass, field
from enum import Enum
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvas
from scipy import signal
from typing_extensions import Annotated

from magicgui import magicgui, register_type, widgets

register_type(float, widget_type="FloatSlider")
register_type(int, widget_type="Slider")

Freq = Annotated[float, {"min": 0.001, "max": 30.0}]
Phase = Annotated[float, {"min": 0.0, "max": 360.0}]
Duty = Annotated[float, {"min": 0.0, "max": 1.0}]
Time = Annotated[float, {"min": 0.01, "max": 100.0}]


@dataclass
class Signal:
    """Constructs a 1D signal.

    As is, this class is not very useful, but one could add callbacks
    or more functionality here

    Parameters
    ----------
    func : callable
        func must take a 'time' array as sole argument and return a 1D array with the
        same size as the input
    duration : float
        the maximum of the input time array
    size : int
        the number of samples in the time array

    """

    func: callable
    duration: Time = 1.0
    size: int = 500
    time: np.ndarray = field(init=False)
    data: np.ndarray = field(init=False)

    def __post_init__(self):
        """Evaluate the function at instantiation time."""
        self.time = np.linspace(0, self.duration, self.size)
        self.data = self.func(self.time)

    def plot(self, ax=None, **kwargs):
        """Plots the data.

        Parameters
        ----------
        ax: matplotlib.axes.Axes instance, default None
           if provided the plot is done on this axes instance.
           If None a new ax is created
        **kwargs: Keyword arguments that are passed on to
            the matplotib ax.plot method

        Returns
        -------
        fig: a matplotlib.figure.Figure instance
        ax: matplotlib.axes.Axes instance
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        ax.plot(self.time, self.data, **kwargs)
        return fig, ax


def sine(
    duration: Time = 10.0, size: int = 500, freq: Freq = 0.5, phase: Phase = 0.0
) -> Signal:
    """Returns a 1D sine wave.

    Parameters
    ----------
    duration: float
       the duration of the signal in seconds
    size: int
        the number of samples in the signal time array
    freq: float
       the frequency of the signal in Hz
    phase: Phase
       the phase of the signal (in degrees)
    """
    sig = Signal(
        duration=duration,
        size=size,
        func=lambda t: np.sin(t * (2 * np.pi * freq) + phase * np.pi / 180),
    )
    return sig


def chirp(
    duration: Time = 10.0,
    size: int = 500,
    f0: float = 1.0,
    t1: Time = 5.0,
    f1: float = 2.0,
    phase: Phase = 0.0,
) -> Signal:
    """Frequency-swept cosine generator.

    See scipy.signal.chirp
    """
    sig = Signal(
        duration=duration,
        size=size,
        func=partial(signal.chirp, f0=f0, t1=t1, f1=f1, phi=phase),
    )
    return sig


def sawtooth(
    duration: Time = 10.0,
    size: int = 500,
    freq: Freq = 1.0,
    width: Duty = 1.0,
    phase: Phase = 0.0,
) -> Signal:
    """Return a periodic sawtooth or triangle waveform.

    See scipy.signal.sawtooth
    """
    sig = Signal(
        duration=duration,
        size=size,
        func=lambda t: signal.sawtooth(
            2 * np.pi * freq * t + phase * np.pi / 180, width=width
        ),
    )
    return sig


def square(
    duration: Time = 10.0, size: int = 500, freq: Freq = 1.0, duty: Duty = 0.5
) -> Signal:
    """Return a periodic sawtooth or triangle waveform.

    See scipy.signal.square
    """
    sig = Signal(
        duration=duration,
        size=size,
        func=lambda t: signal.square(2 * np.pi * freq * t, duty=duty),
    )
    return sig


def on_off(
    duration: Time = 10.0, size: int = 500, t_on: Time = 0.01, t_off: Time = 0.01
) -> Signal:
    """On/Off signal function."""
    data = np.ones(size)
    data[: int(size * t_on / duration)] = -1
    if t_off > 0:
        data[int(size * t_off / duration) :] = -1
    sig = Signal(duration=duration, size=size, func=lambda t: data)
    return sig


WAVEFORMS = {
    "sine": sine,
    "chirp": chirp,
    "sawtooth": sawtooth,
    "square": square,
    "on_off": on_off,
}


class Select(Enum):
    """Enumeration to select signal type."""

    OnOff = "on_off"
    Sine = "sine"
    Chirp = "chirp"
    Sawtooth = "sawtooth"
    Square = "square"


class WaveForm(widgets.Container):
    """Simple waveform generator widget, with plotting."""

    def __init__(self):
        """Creates the widget."""
        super().__init__()
        self.fig, self.ax = plt.subplots()
        self.native.layout().addWidget(FigureCanvas(self.fig))
        self.waveform = sine
        self.controls = None
        self.append(self.signal_widget)
        self.update_controls()
        self.update_graph(sine())

    @magicgui(auto_call=True)
    def signal_widget(self, select: Select = Select.Sine) -> widgets.Container:
        """Waveform selection, from the WAVEFORMS dict."""
        self.waveform = WAVEFORMS[select.value]
        self.update_controls()
        self.update_graph(self.waveform())

    def update_controls(self):
        """Reset controls according to the new function."""
        if self.controls is not None:
            self.remove(self.controls)
        self.controls = magicgui(auto_call=True)(self.waveform)
        self.append(self.controls)
        self.controls.called.connect(self.update_graph)

    def update_graph(self, sig: Signal):
        """Re-plot when a parameter changes.

        Note
        ----
        For big data, this could be slow, maybe `auto_call` should
        not be true in the method above...
        """
        self.ax.cla()
        sig.plot(ax=self.ax)
        self.fig.canvas.draw()


waveform = WaveForm()
waveform.show(run=True)



"""# Optional user choice

Optional user input using a dropdown selection widget.
"""

from typing import Optional

from magicgui import magicgui


# Using optional will add a '----' to the combobox, which returns "None"
@magicgui(path={"choices": ["a", "b"]})
def f(path: Optional[str] = None):
    """Öptional user input function."""
    print(path, type(path))


f.show(run=True)



"""Tests for the Table widget."""

import os
import sys

import pytest

from magicgui.widgets import PushButton, Slider, Table

attr_xfail = pytest.mark.xfail(
    bool(os.getenv("CI")), reason="periodic AttributeError", raises=AttributeError
)

_TABLE_DATA = {
    # column-dict-of-lists
    "list": {"col_1": [1, 4], "col_2": [2, 5], "col_3": [3, 6]},
    # column-dict-of-row-dicts
    "dict": {
        "col_1": {"r1": 1, "r2": 4},
        "col_2": {"r1": 2, "r2": 5},
        "col_3": {"r1": 3, "r2": 6},
    },
    # list-of-lists
    "data": [[1, 2, 3], [4, 5, 6]],
    # Records: List-of-column-dict
    "records": [
        {"col_1": 1, "col_2": 2, "col_3": 3},
        {"col_1": 4, "col_2": 5, "col_3": 6},
    ],
    # 3-tuple of data, index, column
    "tuple": ([[1, 2, 3], [4, 5, 6]], ("r1", "r2"), ("c1", "c2", "c3")),
    # split-dict
    "split": {
        "data": [[1, 2, 3], [4, 5, 6]],
        "index": ("r1", "r2"),
        "columns": ("c1", "c2", "c3"),
    },
}


@pytest.mark.parametrize("key", _TABLE_DATA)
def test_table(key):
    """Test a few ways to input tables."""
    _input = _TABLE_DATA[key]
    table = Table(value=_input)
    if key == "split":
        assert table.value == _input
    if key not in ("tuple", "data"):
        # make sure the output is the same as the input
        assert table.to_dict(key) == _input
        # can also test equality of table widgets
        assert Table(value=table.to_dict(key)) == table
        table.row_headers = ("x", "x")
        table.column_headers = ("a", "a", "b")
        with pytest.warns(UserWarning):
            table.to_dict(key)


def test_constructor():
    """Test various combinations of data, index, column."""
    t = Table(index=["r"], columns=["a", "b"], name="My Table")
    assert t.shape == (1, 2)
    assert t.row_headers == ("r",)
    assert tuple(t) == ("a", "b")
    assert repr(t).split(" at")[0] == "Table(name='My Table', shape=(1, 2)"

    t = Table(index=["a", "b"], columns=["c"])
    assert t.shape == (2, 1)
    assert t.row_headers == ("a", "b")
    assert t.column_headers == ("c",)

    assert Table([1, 2]).shape == (2, 1)  # single list is interpreted as column
    assert Table([[1, 2]]).shape == (1, 2)  # nested lists are rows

    # arguments to constructor override those in a dict value
    t = Table(_TABLE_DATA["dict"], columns=("x", "y", "z"))
    assert t.column_headers == ("x", "y", "z")

    # data and headers can be provided separately
    t = Table([[1, 2, 3], [4, 5, 6]], columns=["col_1", "col_2", "col_3"])
    assert dict(t) == _TABLE_DATA["list"]
    # or together
    assert t == Table(_TABLE_DATA["list"])

    with pytest.warns(UserWarning):
        # more columns than provided... truncated to provided columns
        assert Table([[1, 2]], index=["a"], columns=["v"]).shape == (1, 1)

    with pytest.raises(ValueError):
        # more rows than provided index, raises
        Table([[1], [2]], index=["a"], columns=["v"])
    with pytest.raises(ValueError):
        # same as above
        Table([1, 2], index=["a"], columns=["v"])


def test_adding_deleting_to_empty_table():
    """Test the dict-like api starting with empty table."""
    table = Table()
    assert not any(table.data.to_list())
    assert table.shape == (0, 0)
    # add a new column
    table["c1"] = [1, 2, 3, 4]
    assert table["c1"] == [1, 2, 3, 4]
    assert table.shape == (4, 1)
    # add one with more data
    table["c2"] = [1, 2, 3, 4, 5, 6]
    assert table.shape == (6, 2)
    # add one with less data
    table["c3"] = [1, 2, 3]
    assert table.shape == (6, 3)
    assert table.size == 18
    assert table["c3"][5] is None  # it fills to meet the rows

    assert table.row_headers == (0, 1, 2, 3, 4, 5)
    assert table._get_row(table.row_headers[0]) == [1, 1, 1]
    table._del_row(table.row_headers[0])
    assert table._get_row(table.row_headers[0]) == [2, 2, 2]

    # we can use dict methods
    table.update({"c1": [1, 1]})  # it will clear the extra...
    assert table["c1"] == [1, 1, None, None, None]

    # we can del
    del table["c2"]
    assert "c2" not in table
    with pytest.raises(KeyError):
        table["c2"]  # does not exist

    with pytest.raises(TypeError):
        # not a collection
        table["c5"] = 1  # type: ignore

    with pytest.raises(KeyError):
        del table["c219"]  # does not exist

    table.clear()
    assert table.shape == (0, 0)
    assert not table.column_headers
    assert not table.row_headers


@attr_xfail
def test_orient_index():
    """Test to_dict with orient = 'index' ."""
    table = Table(value=_TABLE_DATA["dict"])
    expected = {
        "r1": {"col_1": 1, "col_2": 2, "col_3": 3},
        "r2": {"col_1": 4, "col_2": 5, "col_3": 6},
    }
    assert table.to_dict("index") == expected

    table = Table(value=_TABLE_DATA["dict"])
    table.row_headers = ("a", "a")
    with pytest.warns(UserWarning):
        table.to_dict("index")

    with pytest.raises(ValueError):
        table.to_dict("notathing")  # type: ignore


def test_table_from_numpy():
    """Test inputting tables from numpy array."""
    np = pytest.importorskip("numpy")
    data = np.arange(12).reshape(4, 3)

    table = Table(value=data)
    assert np.allclose(table.data.to_numpy(), data)


INDICES = (
    1,
    (2, 2),
    (slice(None), 2),
    (slice(None), slice(None)),
    (slice(1, 3), slice(3)),
    (slice(1, None, 3), slice(None, 3)),
    slice(None, None, 2),
    (slice(None), slice(None, None, 2)),
)
VALUES = (
    (7,) * 4,
    6,
    (7,) * 6,
    [[1] * 4] * 6,
    [[1] * 3] * 2,
    [[1] * 3] * 2,
    [[1] * 4] * 3,
    [[1] * 2] * 6,
)


@pytest.mark.parametrize("index", INDICES)
def test_dataview_getitem(index):
    """Test that table.data can be indexed like a numpy array."""
    np = pytest.importorskip("numpy")
    data = np.arange(24).reshape(6, 4)

    table = Table(value=data)
    assert np.allclose(table.data[index], data[index])


@pytest.mark.parametrize("index, value", zip(INDICES, VALUES))
def test_dataview_setitem(index, value):
    """Test that table.data can be indexed like a numpy array."""
    np = pytest.importorskip("numpy")
    data = np.arange(24).reshape(6, 4)

    table = Table(value=data)
    table.data[index] = value
    assert not np.allclose(table.data.to_list(), data)
    data[index] = value
    assert np.allclose(table.data.to_list(), data)


def test_dataview_delitem():
    """Test that table.data can be indexed like a numpy array."""
    _input = _TABLE_DATA["dict"]
    table = Table(value=_input)
    row_keys = table.keys("row")  # also demoing keys views
    col_keys = table.keys("column")  # also demoing keys views
    assert list(row_keys) == ["r1", "r2"]
    assert list(col_keys) == ["col_1", "col_2", "col_3"]
    del table.data[1]
    assert table.to_dict("dict") != _input
    assert list(row_keys) == ["r1"]
    assert list(col_keys) == ["col_1", "col_2", "col_3"]
    del table.data[:, 2]
    assert list(row_keys) == ["r1"]
    assert list(col_keys) == ["col_1", "col_2"]

    with pytest.raises(ValueError):
        del table.data[0, 0]  # cannot delete cells


@attr_xfail
def test_dataview_repr():
    """Test the repr for table.data."""
    table = Table(_TABLE_DATA["dict"], name="My Table")
    assert (
        repr(table.data).split(" at")[0]
        == "<Data for Table(name='My Table', shape=(2, 3)"
    )


def test_table_from_pandas():
    """Test inputting tables from pandas dataframe."""
    pd = pytest.importorskip("pandas", reason="Pandas required for some tables tests")
    df = pd.DataFrame.from_dict(_TABLE_DATA["dict"])
    table = Table(value=df)
    assert pd.DataFrame.equals(table.to_dataframe(), df)


@attr_xfail
def test_orient_series():
    """Test to_dict with orient = 'index' ."""
    pd = pytest.importorskip("pandas", reason="Pandas required for some tables tests")
    table = Table(value=_TABLE_DATA["dict"])
    out = table.to_dict("series")
    assert all(isinstance(s, pd.Series) for s in out.values())


ugly_dict = {
    "col1": {"r1": 8, "r2": 9},
    "col2": {"r1": 10, "r3": 11},
    "col3": {"r7": 12, "r9": 12},
}
ugly_records = [{"col1": 1, "col2": 2}, {"col1": 3}, {"col3": 4, "col2": 5}]


def test_joins_with_pandas():
    """Test that pandas can help with ugly data."""
    pd = pytest.importorskip("pandas", reason="Pandas required for some tables tests")

    t = Table(value=ugly_dict)
    assert t.shape == (5, 3)
    pd.testing.assert_frame_equal(t.to_dataframe(), pd.DataFrame(ugly_dict))

    t2 = Table(value=ugly_records)
    assert t2.shape == (3, 3)
    pd.testing.assert_frame_equal(t2.to_dataframe(), pd.DataFrame(ugly_records))


def test_joins_without_pandas(monkeypatch):
    """Test that we give a useful error with ugly data if pandas is not available."""
    monkeypatch.setitem(sys.modules, "pandas", None)
    with pytest.raises(ValueError) as e:
        Table(value=ugly_dict)
        assert "Install pandas" in str(e)

    with pytest.raises(ValueError) as e:
        Table(value=ugly_records)
        assert "Install pandas" in str(e)


def test_widget_in_table():
    """Test we can put widgets in the table!."""
    table = Table()
    button = PushButton(text="hi")
    slider = Slider(value=50)
    table["a"] = [button, 1, slider, "wow!"]
    assert table["a"] == [button, 1, slider, "wow!"]


def test_view_reprs():
    """Test our custom DictView objects."""
    table = Table(value=_TABLE_DATA["dict"])
    assert repr(table.keys()) == "column_headers(['col_1', 'col_2', 'col_3'])"
    assert repr(table.keys("column")) == "column_headers(['col_1', 'col_2', 'col_3'])"
    assert repr(table.keys("row")) == "row_headers(['r1', 'r2'])"
    assert repr(table.items()) == "table_items(3 columns)"
    assert repr(table.items("column")) == "table_items(3 columns)"
    assert repr(table.items("row")) == "table_items(2 rows)"


def test_row_access_errors():
    """Test various exceptions upon bad access."""
    table = Table(value=_TABLE_DATA["dict"])
    table._set_row("r1", [9, 9, 9])
    assert table._get_row("r1") == [9, 9, 9]
    with pytest.raises(KeyError):
        table._get_row("nonsense")
    with pytest.raises(KeyError):
        table._set_row("nonsense", [1, 2, 3])
    with pytest.raises(KeyError):
        table._del_row("nonsense")
    with pytest.raises(IndexError):
        table._get_rowi(10)

    assert table._assert_col(0) == 0
    with pytest.raises(IndexError):
        table._assert_col(10)


def test_check_new_headers():
    """Check that we get good error messages when setting bad headers."""
    table = Table(value=_TABLE_DATA["dict"])
    with pytest.raises(ValueError) as e:
        table.column_headers = ("a", "b", "c", "d")
        assert "Length mismatch" in str(e)
    with pytest.raises(ValueError) as e:
        table.row_headers = ("a", "b", "c", "d")
        assert "Length mismatch" in str(e)


# these are Qt-specific


def test_copy(qapp):
    from qtpy.QtWidgets import QTableWidgetSelectionRange

    table = Table(value=_TABLE_DATA["data"])
    selrange = QTableWidgetSelectionRange(1, 1, 0, 0)
    table.native.setRangeSelected(selrange, True)
    table.native._copy_to_clipboard()
    assert qapp.clipboard().text() == "1\t2\n4\t5"


def test_paste(qapp):
    from qtpy.QtWidgets import QTableWidgetSelectionRange

    table = Table(value=_TABLE_DATA["data"])
    selrange = QTableWidgetSelectionRange(1, 1, 0, 0)
    table.native.setRangeSelected(selrange, True)
    qapp.clipboard().setText("0\t0\n1\t1")

    table.read_only = True
    table.native._paste_from_clipboard()
    assert table.data.to_list() == [[1, 2, 3], [4, 5, 6]]

    table.read_only = False
    table.native._paste_from_clipboard()
    assert table.data.to_list() == [[0, 0, 3], [1, 1, 6]]


def test_delete(qapp):
    from qtpy.QtWidgets import QTableWidgetSelectionRange

    table = Table(value=_TABLE_DATA["data"])
    selrange = QTableWidgetSelectionRange(1, 1, 0, 0)
    table.native.setRangeSelected(selrange, True)

    table.read_only = True
    assert table.read_only
    table.native._delete_selection()
    assert table.data.to_list() == [[1, 2, 3], [4, 5, 6]]

    table.read_only = False
    assert not table.read_only
    table.native._delete_selection()
    assert table.data.to_list() == [[None, None, 3], [None, None, 6]]


def test_item_delegate(qapp):
    from magicgui.backends._qtpy.widgets import _format_number

    data = ["1.2", "1.23456789", "0.000123", "1234567", "0.0", "1", "s"]
    results = [_format_number(v, 4) for v in data]
    assert results == ["1.2000", "1.2346", "1.230e-04", "1.235e+06", "0.0000", "1", "s"]


def test_row_delete(qapp) -> None:
    table = Table(value=_TABLE_DATA["split"])
    assert table.data.to_list() == [[1, 2, 3], [4, 5, 6]]
    table.delete_row(index=0)
    assert table.data.to_list() == [[4, 5, 6]]
    table.delete_row(header="r2")
    assert table.data.to_list() == []



from __future__ import annotations

import inspect
import os
import sys
import time
from functools import wraps
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterable,
    get_args,
    get_origin,
    overload,
)

from docstring_parser import DocstringParam, parse

if TYPE_CHECKING:
    from typing import TypeVar

    from typing_extensions import ParamSpec

    T = TypeVar("T")
    P = ParamSpec("P")
    C = TypeVar("C", bound=type)


@overload
def debounce(function: Callable[P, T]) -> Callable[P, T | None]: ...


@overload
def debounce(
    *, wait: float = 0.2
) -> Callable[[Callable[P, T]], Callable[P, T | None]]: ...


def debounce(function: Callable[P, T] | None = None, wait: float = 0.2) -> Callable:
    """Postpone function call until `wait` seconds since last invocation."""

    def decorator(fn: Callable[P, T]) -> Callable[P, T | None]:
        from threading import Timer

        _store: dict = {"timer": None, "last_call": 0.0, "args": (), "kwargs": {}}

        @wraps(fn)
        def debounced(*args: P.args, **kwargs: P.kwargs) -> T | None:
            _store["args"] = args
            _store["kwargs"] = kwargs

            def call_it() -> T:
                _store["timer"] = None
                _store["last_call"] = time.time()
                return fn(*_store["args"], **_store["kwargs"])

            if not _store["last_call"]:
                return call_it()

            if _store["timer"] is None:
                time_since_last_call = time.time() - _store["last_call"]
                _store["timer"] = Timer(wait - time_since_last_call, call_it)
                _store["timer"].start()  # type: ignore
            return None

        return debounced

    return decorator if function is None else decorator(function)  # type: ignore


def throttle(t: float) -> Callable[[Callable[P, T]], Callable[P, T | None]]:
    """Prevent a function from being called more than once in `t` seconds."""

    def decorator(f: Callable[P, T]) -> Callable[P, T | None]:
        last = [0.0]

        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T | None:
            if last[0] and (time.time() - last[0] < t):
                return None
            last[0] = time.time()
            return f(*args, **kwargs)

        return wrapper

    return decorator


# modified from appdirs: https://github.com/ActiveState/appdirs
# License: MIT
def user_cache_dir(
    appname: str | None = "magicgui", version: str | None = None
) -> Path:
    r"""Return full path to the user-specific cache dir for this application.

    Typical user cache directories are:
        Mac OS X:   ~/Library/Caches/<AppName>
        Unix:       ~/.cache/<AppName> (XDG default)
        Win XP:     C:\Documents and Settings\<username>\Local Settings\Application Data\<AppName>\Cache  # noqa
        Vista:      C:\Users\<username>\AppData\Local\<AppName>\Cache

    Parameters
    ----------
    appname : str, optional
        Name of application. If None, just the system directory is returned.
        by default "magicgui"
    version : str, optional
        an optional version path element to append to the path. You might want to use
        this if you want multiple versions of your app to be able to run independently.
        If used, this would typically be "<major>.<minor>". Only applied when appname is
        present.

    Returns
    -------
    str
        Full path to the user-specific cache dir for this application.
    """  # noqa: E501
    if sys.platform.startswith("java"):
        import platform

        os_name = platform.java_ver()[3][0]
        if os_name.startswith("Windows"):  # "Windows XP", "Windows 7", etc.
            system = "win32"
        elif os_name.startswith("Mac"):  # "Mac OS X", etc.
            system = "darwin"
        else:  # "Linux", "SunOS", "FreeBSD", etc.
            # Setting this to "linux2" is not ideal, but only Windows or Mac
            # are actually checked for and the rest of the module expects
            # *sys.platform* style strings.
            system = "linux2"
    else:
        system = sys.platform

    home = Path.home()
    if system == "win32":
        _epath = os.getenv("LOCALAPPDATA")
        path = Path(_epath).resolve() if _epath else home / "AppData" / "Local"
        if appname:
            path = path / appname / "Cache"
    elif system == "darwin":
        path = home / "Library" / "Caches"
        if appname:
            path = path / appname
    else:
        _epath = os.getenv("XDG_CACHE_HOME")
        path = Path(_epath) if _epath else home / ".cache"
        if appname:
            path = path / appname
    if appname and version:
        path = path / version
    return path


def _safe_isinstance_tuple(obj: object, superclass: object) -> bool:
    """
    Extracted from `safe_issubclass` to handle checking of generic tuple types.

    It covers following cases:

    1. obj is tuple with ellipsis and superclass is tuple with ellipsis
    2. obj is tuple with ellipsis and superclass is Iterable with specific type
    3. obj is tuple  with same type elements and superclass is tuple with ellipsis

    for other cases it fallback to simple compare types
    """
    obj_args = get_args(obj)
    superclass_args = get_args(superclass)
    superclass_origin = get_origin(superclass)

    if safe_issubclass(superclass_origin, tuple):
        if len(superclass_args) == 2 and superclass_args[1] is Ellipsis:
            if len(obj_args) == 2 and obj_args[1] is Ellipsis:
                # case 3
                return safe_issubclass(obj_args[0], superclass_args[0])
            # case 2
            return all(safe_issubclass(o, superclass_args[0]) for o in obj_args)
        # fallback to simple compare
        return (
                len(obj_args) == len(superclass_args) and
                all(safe_issubclass(o, s) for o, s in zip(obj_args, superclass_args))
        )

    if len(obj_args) == 2 and obj_args[1] is Ellipsis:
        return safe_issubclass(obj_args[0], superclass_args[0])
    return all(safe_issubclass(o, superclass_args[0]) for o in obj_args)


def safe_issubclass(obj: object, superclass: object) -> bool:
    """Safely check if obj is a subclass of superclass."""
    if isinstance(superclass, tuple):
        return any(safe_issubclass(obj, s) for s in superclass)
    obj_origin = get_origin(obj)
    superclass_origin = get_origin(superclass)
    superclass_args = get_args(superclass)
    try:
        if obj_origin is None:
            if superclass_origin is None:
                return issubclass(obj, superclass) # type: ignore
            if not superclass_args:
                return issubclass(obj, superclass_origin)  # type: ignore
            # if obj is not generic type, but superclass is with
            # we can't say anything about it
            return False
        if obj_origin is not None and superclass_origin is None:
            return issubclass(obj_origin, superclass) # type: ignore
        if not issubclass(obj_origin, superclass_origin):  # type: ignore
            return False
        obj_args = get_args(obj)
        if obj_origin is tuple and obj_args:
            return _safe_isinstance_tuple(obj, superclass)

        if not superclass_args:
            return True
        if len(obj_args) != len(superclass_args):
            return False
        return all(safe_issubclass(o, s) for o, s in zip(obj_args, superclass_args))

    except Exception:
        return False


def _param_list_to_str(
    param_list: list[DocstringParam], other_params: list[DocstringParam]
) -> str:
    """Format Parameters section for numpy docstring from list of tuples."""
    out = []
    out += ["Parameters", len("Parameters") * "-"]
    for param in param_list:
        parts = []
        if param.arg_name:
            parts.append(param.arg_name)
        if param.type_name:
            parts.append(param.type_name)
        if not parts:
            continue
        out += [" : ".join(parts)]
        if param.description and param.description.strip():
            out += [" " * 4 + line for line in param.description.split("\n")]
    out += [""]
    return "\n".join(out)


def merge_super_sigs(
    cls: C,
    exclude: Iterable[str] = (
        "widget_type",
        "kwargs",
        "args",
        "kwds",
        "extra",
        "backend_kwargs",
    ),
    module: str = "magicgui.widgets",
) -> C:
    """Merge the signature and kwarg docs from all superclasses, for clearer docs.

    This can be used as a decorator, but must be used with a function call, (even
    if you don't pass any arguments).

    Parameters
    ----------
    cls : Type
        The class being modified
    exclude : tuple, optional
        A list of parameter names to excluded from the merged docs/signature,
        by default ("widget_type", "kwargs", "args", "kwds")
    module : str
        A module name to assign to `cls.__module__`.

    Returns
    -------
    cls : Type
        The modified class (can be used as a decorator)
    """
    params = {}
    param_docs: list[DocstringParam] = []
    other_params: list[DocstringParam] = []
    for sup in inspect.getmro(cls):
        try:
            sig = inspect.signature(sup.__init__)  # type: ignore
        # in some environments `object` or `abc.ABC` will raise ValueError here
        except ValueError:
            continue
        for name, param in sig.parameters.items():
            if name in exclude:
                continue
            params[name] = param

        docstring = getattr(sup, "__doc__", "")
        param_docs += parse(docstring).params
        if "Parameters" in docstring:
            break

    # sphinx_autodoc_typehints isn't removing the type annotations from the signature
    # so we do it manually when building documentation.
    if sys.argv[-2:] == ["build", "docs"]:  # if building docs
        params = {
            k: v.replace(annotation=inspect.Parameter.empty) for k, v in params.items()
        }

    cls.__init__.__signature__ = inspect.Signature(  # type: ignore
        sorted(params.values(), key=lambda x: x.kind)
    )
    param_docs = [p for p in param_docs if p.arg_name not in exclude]
    cls.__doc__ = (cls.__doc__ or "").split("Parameters")[0].rstrip() + "\n\n"
    cls.__doc__ += _param_list_to_str(param_docs, other_params)
    # this makes docs linking work... but requires that all of these be in __init__
    cls.__module__ = module
    return cls


# def mergedoc(
#     cls: type,
#     exclude: Iterable[str] = (
#         "widget_type",
#         "kwargs",
#         "args",
#         "kwds",
#         "extra",
#         "backend_kwargs",
#     ),
# ):
#     from textwrap import indent

#     from griffe.dataclasses import Docstring
#     from griffe.docstrings.dataclasses import DocstringParameter, DocstringSection
#     from griffe.docstrings.numpy import parse

#     super_docs: dict[str, list[DocstringSection]] = {}
#     for sup in cls.__mro__:
#         docstring = getattr(sup, "__doc__", None)
#         if docstring:
#             super_docs[sup.__name__] = parse(Docstring(docstring))

#     seen = set(exclude)
#     for base, sections in super_docs.items():
#         print(">>", base)
#         for section in sections:
#             if section.kind.name == "parameters":
#                 for p in section.value:
#                     if p.name in seen:
#                         # continue
#                         ...
#                     p = cast(DocstringParameter, p)
#                     desc = indent(p.description, ' ' * 4)
#                     line = f"{p.name} : {p.annotation}\n{desc}"
#                     seen.add(p.name)
#                     print(line)



from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Sequence, Union
from unittest.mock import Mock

import pytest
from typing_extensions import Annotated, get_args

from magicgui import magicgui, register_type, type_map, type_registered, types, widgets
from magicgui._type_resolution import resolve_single_type
from magicgui.type_map._type_map import _RETURN_CALLBACKS

if TYPE_CHECKING:
    import numpy


def test_forward_refs():
    """Test that forward refs parameter annotations get resolved."""

    @magicgui
    def testA(x: "tests.MyInt" = "1"):  # type: ignore  # noqa
        pass

    @magicgui
    def testB(x="1"):
        pass

    # because tests.MyInt is a subclass of int, it will be shown as a SpinBox
    assert isinstance(testA.x, widgets.SpinBox)
    # whereas without the forward ref type annotation, it would have been a LineEdit
    assert isinstance(testB.x, widgets.LineEdit)

    with pytest.raises(ImportError) as err:
        # bad forward ref
        @magicgui
        def testA(x: "testsd.MyInt" = "1"):  # type: ignore  # noqa
            pass

    assert "Magicgui could not resolve ForwardRef" in str(err.value)


@pytest.mark.parametrize(
    "cls, string", [("LineEdit", "str"), ("SpinBox", "int"), ("FloatSpinBox", "float")]
)
def test_pick_widget_builtins_forward_refs(cls, string):
    wdg = type_map.get_widget_class(annotation=string)[0]
    assert wdg.__name__ == cls


@pytest.mark.parametrize(
    "hint, expected_wdg",
    [
        (Annotated[int, {"min": 8, "max": 9}], widgets.SpinBox),
        (
            Annotated[float, {"widget_type": "FloatSlider", "step": 9}],
            widgets.FloatSlider,
        ),
        (
            Annotated[Annotated[int, {"widget_type": "Slider"}], {"max": 10}],
            widgets.Slider,
        ),
    ],
)
def test_annotated_types(hint, expected_wdg):
    wdg, options = type_map.get_widget_class(annotation=hint)
    assert wdg is expected_wdg
    for k, v in get_args(hint)[1].items():
        if k != "widget_type":
            assert options[k] == v


def test_forward_refs_return_annotation():
    """Test that forward refs return annotations get resolved."""

    @magicgui
    def testA() -> int:
        return 1

    @magicgui
    def testB() -> "tests.MyInt":  # type: ignore  # noqa
        return 1

    from tests import MyInt

    results = []
    register_type(MyInt, return_callback=lambda *x: results.append(x))

    testA()
    assert not results

    testB()
    gui, result, return_annotation = results[0]
    assert isinstance(gui, widgets.FunctionGui)
    assert result == 1
    # the forward ref has been resolved
    assert return_annotation is MyInt


def test_pathlike_annotation():
    import pathlib

    @magicgui(fn={"mode": "r"})
    def widget(fn: types.PathLike):
        print(fn)

    assert isinstance(widget.fn, widgets.FileEdit)
    assert widget.fn.mode is types.FileDialogMode.EXISTING_FILE

    # an equivalent union also works
    @magicgui(fn={"mode": "rm"})
    def widget2(fn: Union[bytes, pathlib.Path, str]):
        print(fn)

    assert isinstance(widget2.fn, widgets.FileEdit)
    assert widget2.fn.mode is types.FileDialogMode.EXISTING_FILES


def test_optional_type():
    @magicgui(x={"choices": ["a", "b"]})
    def widget(x: Optional[str] = None):
        ...

    assert isinstance(widget.x, widgets.ComboBox)
    assert widget.x.value is None
    assert None in widget.x.choices


def test_widget_options():
    """Test bugfix: widget options shouldn't persist to next widget."""
    E = Enum("E", ["a", "b", "c"])
    choice1 = widgets.create_widget(annotation=E)
    choice2 = widgets.create_widget(annotation=Optional[E])
    choice3 = widgets.create_widget(annotation=E)
    assert choice1._nullable is choice3._nullable is False
    assert choice2._nullable is True


def test_nested_forward_refs():
    resolved = resolve_single_type(Optional['List["numpy.ndarray"]'])

    from typing import List

    import numpy as np

    assert resolved == Optional[List[np.ndarray]]


def test_type_registered():
    assert isinstance(widgets.create_widget(annotation=Path), widgets.FileEdit)
    with type_registered(Path, widget_type=widgets.LineEdit):
        assert isinstance(widgets.create_widget(annotation=Path), widgets.LineEdit)
    assert isinstance(widgets.create_widget(annotation=Path), widgets.FileEdit)
    assert isinstance(widgets.create_widget(annotation=Sequence[Path]), widgets.FileEdit)


def test_type_registered_callbacks():
    @magicgui
    def func(a: int) -> int:
        return a

    assert not _RETURN_CALLBACKS[int]
    mock = Mock()
    func(1)
    mock.assert_not_called()

    cb = lambda g, v, r: mock(v)  # noqa
    cb2 = lambda g, v, r: None  # noqa

    with type_registered(int, return_callback=cb):
        func(2)
        mock.assert_called_once_with(2)
        mock.reset_mock()
        assert _RETURN_CALLBACKS[int] == [cb]
        register_type(int, return_callback=cb2)
        assert _RETURN_CALLBACKS[int] == [cb, cb2]

    func(3)
    mock.assert_not_called()
    assert _RETURN_CALLBACKS[int] == [cb2]


def test_type_registered_warns():
    """Test that type_registered warns if the type was changed during context."""
    assert isinstance(widgets.create_widget(annotation=Path), widgets.FileEdit)
    with pytest.warns(UserWarning, match="Type definition changed during context"):
        with type_registered(Path, widget_type=widgets.LineEdit):
            assert isinstance(widgets.create_widget(annotation=Path), widgets.LineEdit)
            register_type(Path, widget_type=widgets.TextEdit)
            assert isinstance(widgets.create_widget(annotation=Path), widgets.TextEdit)
    assert isinstance(widgets.create_widget(annotation=Path), widgets.FileEdit)


def test_type_registered_optional_callbacks():
    assert not _RETURN_CALLBACKS[int]
    assert not _RETURN_CALLBACKS[Optional[int]]

    @magicgui
    def func1(a: int) -> int:
        return a

    @magicgui
    def func2(a: int) -> Optional[int]:
        return a

    mock1 = Mock()
    mock2 = Mock()
    mock3 = Mock()

    register_type(int, return_callback=mock2)

    with type_registered(Optional[int], return_callback=mock1):
        func1(1)
        mock1.assert_called_once_with(func1, 1, int)
        mock1.reset_mock()
        func2(2)
        mock1.assert_called_once_with(func2, 2, Optional[int])
        mock1.reset_mock()
        mock2.assert_called_once_with(func1, 1, int)
        assert _RETURN_CALLBACKS[int] == [mock2, mock1]
        assert _RETURN_CALLBACKS[Optional[int]] == [mock1]
        register_type(Optional[int], return_callback=mock3)
        assert _RETURN_CALLBACKS[Optional[int]] == [mock1, mock3]

    assert _RETURN_CALLBACKS[Optional[int]] == [mock3]
    assert _RETURN_CALLBACKS[int] == [mock2, mock3]


def test_pick_widget_literal():
    from typing import Literal

    cls, options = type_map.get_widget_class(
        annotation=Annotated[Literal["a", "b"], {"widget_type": "RadioButtons"}]
    )
    assert cls == widgets.RadioButtons
    assert set(options["choices"]) == {"a", "b"}


def test_redundant_annotation() -> None:
    # just shouldn't raise
    @magicgui
    def f(a: Annotated[List[int], {"annotation": List[int]}]):
        pass



"""# Nested progress bars

Example using nested progress bars in magicgui.

"""

import random
from time import sleep

from magicgui import magicgui
from magicgui.tqdm import tqdm, trange

# if magicui.tqdm.tqdm or trange are used outside of a @magicgui function, (such as in
# interactive use in IPython), then they fall back to the standard terminal output


# If use inside of a magicgui-decorated function
# a progress bar widget will be added to the magicgui container
@magicgui(call_button=True, layout="vertical")
def long_function(
    steps=10, repeats=4, choices="ABCDEFGHIJKLMNOP12345679", char="", delay=0.05
):
    """Long running computation with nested iterators."""
    # trange and tqdm accept all the kwargs from tqdm itself, as well as any
    # valid kwargs for magicgui.widgets.ProgressBar, (such as "label")
    for _r in trange(repeats, label="repeats"):
        letters = [random.choice(choices) for _ in range(steps)]
        # `tqdm`, like `tqdm`, accepts any iterable
        # this progress bar is nested and will be run & reset multiple times
        for letter in tqdm(letters, label="steps"):
            long_function.char.value = letter
            sleep(delay)


long_function.show(run=True)



"""Functions in this module are responsible for mapping type annotations to widgets."""

from __future__ import annotations

import datetime
import inspect
import itertools
import os
import pathlib
import sys
import types
import warnings
from collections import defaultdict
from contextlib import contextmanager
from enum import EnumMeta
from typing import (
    Any,
    Callable,
    Dict,
    ForwardRef,
    Iterator,
    Literal,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from typing_extensions import Annotated, get_args, get_origin

from magicgui import widgets
from magicgui._type_resolution import resolve_single_type
from magicgui._util import safe_issubclass
from magicgui.types import PathLike, ReturnCallback, Undefined, _Undefined
from magicgui.widgets.protocols import WidgetProtocol, assert_protocol

__all__: list[str] = ["register_type", "get_widget_class"]


# redefining these here for the sake of sphinx autodoc forward refs
WidgetClass = Union[Type[widgets.Widget], Type[WidgetProtocol]]
WidgetRef = Union[str, WidgetClass]
WidgetTuple = Tuple[WidgetRef, Dict[str, Any]]


class MissingWidget(RuntimeError):
    """Raised when a backend widget cannot be found."""


_RETURN_CALLBACKS: defaultdict[type, list[ReturnCallback]] = defaultdict(list)
_TYPE_DEFS: dict[type, WidgetTuple] = {}


_SIMPLE_ANNOTATIONS = {
    PathLike: widgets.FileEdit,
}

_SIMPLE_TYPES = {
    bool: widgets.CheckBox,
    int: widgets.SpinBox,
    float: widgets.FloatSpinBox,
    str: widgets.LineEdit,
    pathlib.Path: widgets.FileEdit,
    datetime.time: widgets.TimeEdit,
    datetime.timedelta: widgets.TimeEdit,
    datetime.date: widgets.DateEdit,
    datetime.datetime: widgets.DateTimeEdit,
    range: widgets.RangeEdit,
    slice: widgets.SliceEdit,
    Sequence[pathlib.Path]: widgets.FileEdit,
    tuple: widgets.TupleEdit,
    Sequence: widgets.ListEdit,
    os.PathLike: widgets.FileEdit,
}

_ADDITIONAL_KWARGS: dict[type, dict[str, Any]] = {
    Sequence[pathlib.Path]: {"mode": "rm"}
}

def match_type(type_: Any, default: Any | None = None) -> WidgetTuple | None:
    """Check simple type mappings."""
    if type_ in _SIMPLE_ANNOTATIONS:
        return _SIMPLE_ANNOTATIONS[type_], {}

    if type_ is widgets.ProgressBar:
        return widgets.ProgressBar, {"bind": lambda widget: widget, "visible": True}

    if type_ in _SIMPLE_TYPES:
        return _SIMPLE_TYPES[type_], _ADDITIONAL_KWARGS.get(type_, {})
    for key in _SIMPLE_TYPES.keys():
        if safe_issubclass(type_, key):
            return _SIMPLE_TYPES[key], _ADDITIONAL_KWARGS.get(key, {})

    if type_ in (types.FunctionType,):
        return widgets.FunctionGui, {"function": default}

    origin = get_origin(type_) or type_
    choices, nullable = _literal_choices(type_)
    if choices is not None:  # it's a Literal type
        return widgets.ComboBox, {"choices": choices, "nullable": nullable}

    if safe_issubclass(origin, Set):
        for arg in get_args(type_):
            if get_origin(arg) is Literal:
                return widgets.Select, {"choices": get_args(arg)}

    pint = sys.modules.get("pint")
    if pint and safe_issubclass(origin, pint.Quantity):
        return widgets.QuantityEdit, {}

    return None


_SIMPLE_RETURN_TYPES = [
    bool,
    int,
    float,
    str,
    pathlib.Path,
    datetime.time,
    datetime.date,
    datetime.datetime,
    range,
    slice,
]


def match_return_type(type_: Any) -> WidgetTuple | None:
    """Check simple type mappings for result widgets."""
    if type_ in _SIMPLE_TYPES:
        return widgets.LineEdit, {"gui_only": True}

    if type_ is widgets.Table:
        return widgets.Table, {}

    table_types = [
        resolve_single_type(x) for x in ("pandas.DataFrame", "numpy.ndarray")
    ]

    if any(
        safe_issubclass(type_, tt)
        for tt in table_types
        if not isinstance(tt, ForwardRef)
    ):
        return widgets.Table, {}

    return None


def _is_none_type(type_: Any) -> bool:
    return any(type_ is x for x in {None, type(None), Literal[None]})


def _type_optional(
    default: Any = Undefined,
    annotation: type[Any] | _Undefined = Undefined,
) -> tuple[Any, bool]:
    type_ = annotation
    if annotation in (Undefined, None, inspect.Parameter.empty):
        if default is not Undefined:
            type_ = type(default)
    else:
        try:
            type_ = resolve_single_type(annotation)
        except (NameError, ImportError) as e:
            raise type(e)(f"Magicgui could not resolve {annotation}: {e}") from e

    # look for Optional[Type], which manifests as Union[Type, None]
    nullable = default is None
    if type_ is not Undefined:
        args = get_args(type_)
        for arg in args:
            if _is_none_type(arg) or arg is Any or arg is object:
                nullable = True
                if len(args) == 2:
                    type_ = next(i for i in args if i is not arg)
                break

    return type_, nullable


def _literal_choices(annotation: Any) -> tuple[list | None, bool]:
    """Return choices and nullable for a Literal type.

    if annotation is not a Literal type, returns (None, False)
    """
    origin = get_origin(annotation) or annotation
    choices: list | None = None
    nullable = False
    if origin is Literal:
        choices = []
        for choice in get_args(annotation):
            if choice is None:
                nullable = True
            else:
                choices.append(choice)
    return choices, nullable


def _pick_widget_type(
    value: Any = Undefined,
    annotation: Any = Undefined,
    options: dict | None = None,
    is_result: bool = False,
    raise_on_unknown: bool = True,
) -> WidgetTuple:
    """Pick the appropriate widget type for ``value`` with ``annotation``."""
    annotation, options_ = _split_annotated_type(annotation)
    options = {**options_, **(options or {})}
    choices = options.get("choices")

    if is_result and annotation is inspect.Parameter.empty:
        annotation = str

    if (
        value is Undefined
        and annotation in (Undefined, inspect.Parameter.empty)
        and not choices
        and "widget_type" not in options
    ):
        return widgets.EmptyWidget, {"visible": False, **options}

    type_, optional = _type_optional(value, annotation)
    options.setdefault("nullable", optional)
    choices = choices or (isinstance(type_, EnumMeta) and type_)
    literal_choices, nullable = _literal_choices(annotation)
    if literal_choices is not None:
        choices = literal_choices
        options["nullable"] = nullable

    if "widget_type" in options:
        widget_type = options.pop("widget_type")
        if choices:
            if widget_type == "RadioButton":
                widget_type = "RadioButtons"
                warnings.warn(
                    f"widget_type of 'RadioButton' (with dtype {type_}) is"
                    " being coerced to 'RadioButtons' due to choices or Enum type.",
                    stacklevel=2,
                )
            options.setdefault("choices", choices)
        return widget_type, options

    # look for subclasses
    for registered_type in _TYPE_DEFS:
        if type_ == registered_type or safe_issubclass(type_, registered_type):
            cls_, opts = _TYPE_DEFS[registered_type]
            return cls_, {**options, **opts}

    if is_result:
        widget_type_ = match_return_type(type_)
        if widget_type_:
            cls_, opts = widget_type_
            return cls_, {**opts, **options}
        # Chosen for backwards/test compatibility
        return widgets.LineEdit, {"gui_only": True}

    if choices:
        options["choices"] = choices
        wdg = widgets.Select if options.get("allow_multiple") else widgets.ComboBox
        return wdg, options

    widget_type_ = match_type(type_, value)
    if widget_type_:
        cls_, opts = widget_type_
        return cls_, {**opts, **options}

    if raise_on_unknown:
        raise ValueError(
            f"No widget found for type {type_} and annotation {annotation!r}"
        )

    options["visible"] = False
    return widgets.EmptyWidget, options


def _split_annotated_type(annotation: Any) -> tuple[Any, dict]:
    """Split an Annotated type into its base type and options dict."""
    if get_origin(annotation) is not Annotated:
        return annotation, {}

    type_, meta_, *_ = get_args(annotation)

    try:
        meta = dict(meta_)
    except TypeError:
        meta = {}

    return type_, meta


def get_widget_class(
    value: Any = Undefined,
    annotation: Any = Undefined,
    options: dict | None = None,
    is_result: bool = False,
    raise_on_unknown: bool = True,
) -> tuple[WidgetClass, dict]:
    """Return a [Widget][magicgui.widgets.Widget] subclass for the `value`/`annotation`.

    Parameters
    ----------
    value : Any, optional
        A python value.  Will be used to determine the widget type if an ``annotation``
        is not explicitly provided by default None
    annotation : Optional[Type], optional
        A type annotation, by default None
    options : dict, optional
        Options to pass when constructing the widget, by default {}
    is_result : bool, optional
        Identifies whether the returned widget should be tailored to
        an input or to an output.
    raise_on_unknown : bool, optional
        Raise exception if no widget is found for the given type, by default True

    Returns
    -------
    Tuple[WidgetClass, dict]
        The WidgetClass, and dict that can be used for params. dict
        may be different than the options passed in.
    """
    widget_type, options_ = _pick_widget_type(
        value, annotation, options, is_result, raise_on_unknown
    )

    if isinstance(widget_type, str):
        widget_class = _import_wdg_class(widget_type)
    else:
        widget_class = widget_type

    if not safe_issubclass(widget_class, widgets.bases.Widget):
        assert_protocol(widget_class, WidgetProtocol)

    return widget_class, options_


def _import_wdg_class(class_name: str) -> WidgetClass:
    import importlib

    # import from magicgui widgets if not explicitly namespaced
    if "." not in class_name:
        class_name = f"magicgui.widgets.{class_name}"

    mod_name, name = class_name.rsplit(".", 1)
    mod = importlib.import_module(mod_name)
    return cast(WidgetClass, getattr(mod, name))


def _validate_return_callback(func: Callable) -> None:
    try:
        sig = inspect.signature(func)
        # the signature must accept three arguments
        sig.bind(1, 2, 3)  # (gui, result, return_type)
    except TypeError as e:
        raise TypeError(f"object {func!r} is not a valid return callback: {e}") from e


_T = TypeVar("_T", bound=type)


def _register_type_callback(
    resolved_type: _T,
    return_callback: ReturnCallback | None = None,
) -> list[type]:
    modified_callbacks = []
    if return_callback is None:
        return []
    _validate_return_callback(return_callback)
    # if the type is a Union, add the callback to all of the types in the union
    # (except NoneType)
    if get_origin(resolved_type) is Union:
        for type_per in _generate_union_variants(resolved_type):
            if return_callback not in _RETURN_CALLBACKS[type_per]:
                _RETURN_CALLBACKS[type_per].append(return_callback)
                modified_callbacks.append(type_per)

        for t in get_args(resolved_type):
            if not _is_none_type(t) and return_callback not in _RETURN_CALLBACKS[t]:
                _RETURN_CALLBACKS[t].append(return_callback)
                modified_callbacks.append(t)
    elif return_callback not in _RETURN_CALLBACKS[resolved_type]:
        _RETURN_CALLBACKS[resolved_type].append(return_callback)
        modified_callbacks.append(resolved_type)
    return modified_callbacks


def _register_widget(
    resolved_type: _T,
    widget_type: WidgetRef | None = None,
    **options: Any,
) -> WidgetTuple | None:
    _options = cast(dict, options)

    previous_widget = _TYPE_DEFS.get(resolved_type)

    if "choices" in _options:
        _TYPE_DEFS[resolved_type] = (widgets.ComboBox, _options)
        if widget_type is not None:
            warnings.warn(
                "Providing `choices` overrides `widget_type`. Categorical widget "
                f"will be used for type {resolved_type}",
                stacklevel=2,
            )
    elif widget_type is not None:
        if not isinstance(widget_type, (str, WidgetProtocol)) and not (
            inspect.isclass(widget_type) and issubclass(widget_type, widgets.Widget)
        ):
            raise TypeError(
                '"widget_type" must be either a string, WidgetProtocol, or '
                "Widget subclass"
            )
        _TYPE_DEFS[resolved_type] = (widget_type, _options)
    elif "bind" in _options:
        # if we're binding a value to this parameter, it doesn't matter what type
        # of ValueWidget is used... it usually won't be shown
        _TYPE_DEFS[resolved_type] = (widgets.EmptyWidget, _options)
    return previous_widget


@overload
def register_type(
    type_: _T,
    *,
    widget_type: WidgetRef | None = None,
    return_callback: ReturnCallback | None = None,
    **options: Any,
) -> _T: ...


@overload
def register_type(
    type_: Literal[None] | None = None,
    *,
    widget_type: WidgetRef | None = None,
    return_callback: ReturnCallback | None = None,
    **options: Any,
) -> Callable[[_T], _T]: ...


def register_type(
    type_: _T | None = None,
    *,
    widget_type: WidgetRef | None = None,
    return_callback: ReturnCallback | None = None,
    **options: Any,
) -> _T | Callable[[_T], _T]:
    """Register a ``widget_type`` to be used for all parameters with type ``type_``.

    Note: registering a Union (or Optional) type effectively registers all types in
    the union with the arguments.

    Parameters
    ----------
    type_ : type
        The type for which a widget class or return callback will be provided.
    widget_type : WidgetRef, optional
        A widget class from the current backend that should be used whenever ``type_``
        is used as the type annotation for an argument in a decorated function,
        by default None
    return_callback: callable, optional
        If provided, whenever ``type_`` is declared as the return type of a decorated
        function, ``return_callback(widget, value, return_type)`` will be called
        whenever the decorated function is called... where ``widget`` is the Widget
        instance, and ``value`` is the return value of the decorated function.
    options
        key value pairs where the keys are valid `dict`

    Raises
    ------
    ValueError
        If none of `widget_type`, `return_callback`, `bind` or `choices` are provided.
    """
    if all(
        x is None
        for x in [
            return_callback,
            options.get("bind"),
            options.get("choices"),
            widget_type,
        ]
    ):
        raise ValueError(
            "At least one of `widget_type`, `return_callback`, `bind` or `choices` "
            "must be provided."
        )

    def _deco(type__: _T) -> _T:
        resolved_type = resolve_single_type(type__)
        _register_type_callback(resolved_type, return_callback)
        _register_widget(resolved_type, widget_type, **options)
        return type__

    return _deco if type_ is None else _deco(type_)


@contextmanager
def type_registered(
    type_: _T,
    *,
    widget_type: WidgetRef | None = None,
    return_callback: ReturnCallback | None = None,
    **options: Any,
) -> Iterator[None]:
    """Context manager that temporarily registers a widget type for a given `type_`.

    When the context is exited, the previous widget type associations for `type_` is
    restored.

    Parameters
    ----------
    type_ : _T
        The type for which a widget class or return callback will be provided.
    widget_type : Optional[WidgetRef]
        A widget class from the current backend that should be used whenever ``type_``
        is used as the type annotation for an argument in a decorated function,
        by default None
    return_callback: Optional[callable]
        If provided, whenever ``type_`` is declared as the return type of a decorated
        function, ``return_callback(widget, value, return_type)`` will be called
        whenever the decorated function is called... where ``widget`` is the Widget
        instance, and ``value`` is the return value of the decorated function.
    options
        key value pairs where the keys are valid `dict`
    """
    resolved_type = resolve_single_type(type_)

    # store any previous widget_type and options for this type

    revert_list = _register_type_callback(resolved_type, return_callback)
    prev_type_def = _register_widget(resolved_type, widget_type, **options)

    new_type_def: WidgetTuple | None = _TYPE_DEFS.get(resolved_type, None)
    try:
        yield
    finally:
        # restore things to before the context
        if return_callback is not None:  # this if is only for mypy
            for return_callback_type in revert_list:
                _RETURN_CALLBACKS[return_callback_type].remove(return_callback)

        if _TYPE_DEFS.get(resolved_type, None) is not new_type_def:
            warnings.warn("Type definition changed during context", stacklevel=2)

        if prev_type_def is not None:
            _TYPE_DEFS[resolved_type] = prev_type_def
        else:
            _TYPE_DEFS.pop(resolved_type, None)


def type2callback(type_: type) -> list[ReturnCallback]:
    """Return any callbacks that have been registered for ``type_``.

    Parameters
    ----------
    type_ : type
        The type_ to look up.

    Returns
    -------
    list of callable
        Where a return callback accepts two arguments (gui, value) and does something.
    """
    if type_ is inspect.Parameter.empty:
        return []

    # look for direct hits ...
    # if it's an Optional, we need to look for the type inside the Optional
    type_ = resolve_single_type(type_)
    if type_ in _RETURN_CALLBACKS:
        return _RETURN_CALLBACKS[type_]

    # look for subclasses
    for registered_type in _RETURN_CALLBACKS:  # sourcery skip: use-next
        if safe_issubclass(type_, registered_type):
            return _RETURN_CALLBACKS[registered_type]
    return []


def _generate_union_variants(type_: Any) -> Iterator[type]:
    type_args = get_args(type_)
    for i in range(2, len(type_args) + 1):
        for per in itertools.combinations(type_args, i):
            yield cast(type, Union[per])



"""# Multiple selection widget

A selection widget allowing multiple selections by the user.
"""

from magicgui import magicgui


@magicgui(
    pick_some={
        "choices": ("first", "second", "third", "fourth"),
        "allow_multiple": True,
    }
)
def my_widget(pick_some=("first")):
    """Dropdown selection function."""
    print("you selected", pick_some)


my_widget.show(run=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from magicgui.application import AppRef, use_app
from magicgui.types import Undefined
from magicgui.widgets import bases, protocols

from ._widget import Widget

if TYPE_CHECKING:
    import inspect


def create_widget(
    value: Any = Undefined,
    annotation: Any | None = None,
    name: str = "",
    param_kind: str | inspect._ParameterKind = "POSITIONAL_OR_KEYWORD",
    label: str | None = None,
    gui_only: bool = False,
    app: AppRef = None,
    widget_type: str | type[protocols.WidgetProtocol] | None = None,
    options: dict | None = None,
    is_result: bool = False,
    raise_on_unknown: bool = True,
) -> Widget:
    """Create and return appropriate widget subclass.

    This factory function can be used to create a widget appropriate for the
    provided ``value`` and/or ``annotation`` provided. See
    [Type Mapping Docs](../../type_map.md) for details on how the widget type is
    determined from type annotations.

    Parameters
    ----------
    value : Any, optional
        The starting value for the widget, by default ``None``
    annotation : Any, optional
        The type annotation for the parameter represented by the widget, by default
        ``None``.
    name : str, optional
        The name of the parameter represented by this widget. by default ``""``
    param_kind : str, optional
        The :attr:`inspect.Parameter.kind` represented by this widget.  Used in building
        signatures from multiple widgets, by default "``POSITIONAL_OR_KEYWORD``"
    label : str
        A string to use for an associated Label widget (if this widget is being
        shown in a [`Container`][magicgui.widgets.Container] widget, and labels are on).
        By default, ``name`` will be used. Note: ``name`` refers the name of the
        parameter, as might be used in a signature, whereas label is just the label
        for that widget in the GUI.
    gui_only : bool, optional
        Whether the widget should be considered "only for the gui", or if it should
        be included in any widget container signatures, by default False
    app : str, optional
        The backend to use, by default ``None``
    widget_type : str or Type[WidgetProtocol] or None
        A class implementing a widget protocol or a string with the name of a
        magicgui widget type (e.g. "Label", "PushButton", etc...).
        If provided, this widget type will be used instead of the type
        autodetermined from ``value`` and/or ``annotation`` above.
    options : dict, optional
        Dict of options to pass to the Widget constructor, by default dict()
    is_result : boolean, optional
        Whether the widget belongs to an input or an output. By default, an input
        is assumed.
    raise_on_unknown : bool, optional
        Raise exception if no widget is found for the given type, by default True

    Returns
    -------
    Widget
        An instantiated widget subclass

    Raises
    ------
    TypeError
        If the provided or autodetected ``widget_type`` does not implement any known
        [widget protocols](../protocols.md)

    Examples
    --------
    ```python
    from magicgui.widgets import create_widget

    # create a widget from a string value
    wdg = create_widget(value="hello world")
    assert wdg.value == "hello world"

    # create a widget from a string annotation
    wdg = create_widget(annotation=str)
    assert wdg.value == ""
    ```
    """
    options_ = options.copy() if options is not None else {}
    kwargs = {
        "value": value,
        "annotation": annotation,
        "name": name,
        "label": label,
        "gui_only": gui_only,
    }

    assert use_app(app).native
    if isinstance(widget_type, protocols.WidgetProtocol):
        wdg_class = widget_type
    else:
        from magicgui.type_map import get_widget_class

        if widget_type:
            options_["widget_type"] = widget_type
        # special case parameters named "password" with annotation of str
        if (
            not options_.get("widget_type")
            and (name or "").lower() == "password"
            and annotation is str
        ):
            options_["widget_type"] = "Password"

        wdg_class, opts = get_widget_class(
            value, annotation, options_, is_result, raise_on_unknown
        )

        if issubclass(wdg_class, Widget):
            widget = wdg_class(**{**kwargs, **opts, **options_})
            if param_kind:
                widget.param_kind = param_kind  # type: ignore
            return widget

    # pick the appropriate subclass for the given protocol
    # order matters
    for p in ("Categorical", "Ranged", "Button", "Value", ""):
        prot = getattr(protocols, f"{p}WidgetProtocol")
        if isinstance(wdg_class, prot):
            options_ = kwargs.pop("options", None)
            cls = getattr(bases, f"{p}Widget")
            widget = cls(**{**kwargs, **(options_ or {}), "widget_type": wdg_class})
            if param_kind:
                widget.param_kind = param_kind  # type: ignore
            return widget

    raise TypeError(f"{wdg_class!r} does not implement any known widget protocols")



from __future__ import annotations

import dataclasses as dc
import sys
import warnings
from dataclasses import dataclass, field
from functools import lru_cache
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Literal,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Annotated, TypeGuard, get_args, get_origin

from magicgui.types import JsonStringFormats, Undefined, _Undefined

if TYPE_CHECKING:
    from typing import Mapping, Protocol

    import attrs
    import pydantic
    from annotated_types import BaseMetadata
    from attrs import Attribute
    from pydantic.fields import FieldInfo, ModelField

    from magicgui.widgets.bases import ContainerWidget, ValueWidget

    class HasAttrs(Protocol):
        """Protocol for objects that have an ``attrs`` attribute."""

        __attrs_attrs__: tuple[attrs.Attribute, ...]


__all__ = ["build_widget", "get_ui_fields", "UiField"]

SLOTS = {"slots": True} if sys.version_info >= (3, 10) else {}
T = TypeVar("T")


@dataclass(frozen=True, **SLOTS)
class UiField(Generic[T]):
    """Metadata about a specific widget in a GUI."""

    def __post_init__(self) -> None:
        """Coerce Optional[...] to nullable and remove it from the type."""
        if get_origin(self.type) is Union:
            args = get_args(self.type)
            nonnull = tuple(a for a in args if a is not type(None))
            if len(nonnull) < len(args):
                # object.__setattr__ because we are using a frozen dataclass
                object.__setattr__(self, "_original_annotation", self.type)
                object.__setattr__(self, "type", Union[nonnull])
                object.__setattr__(self, "nullable", True)

    name: str | None = field(
        default=None,
        metadata={
            "description": "The name of the field.  This differs from `title` in that "
            "refers to the python name used to refer to this value. e.g. the parameter "
            "name, or field name in a dataclass"
        },
    )

    # Basic Meta-Data Annotations vocabulary
    # https://json-schema.org/draft/2020-12/json-schema-validation.html#name-a-vocabulary-for-basic-meta
    title: str | None = field(
        default=None,
        metadata={
            "description": "A short title for the field.  If not provided, "
            "the `name` will be used.",
            "aliases": ["label", "text", "button_text"],
        },
    )
    description: str | None = field(
        default=None,
        metadata={"description": "A description of the field.", "aliases": ["tooltip"]},
    )
    default: T = field(  # type: ignore
        default=Undefined,
        metadata={
            "description": "The default value for the field.",
            "aliases": ["value"],
        },
    )
    # NOTE: this does not have an analog in JSON Schema
    default_factory: Callable[[], T] | None = field(
        default=None,
        metadata={
            "description": "A callable that returns the default value of the field."
        },
    )
    # NOTE: this does not have an analog in JSON Schema
    nullable: bool | None = field(
        default=None,
        metadata={
            "description": "Whether the field is nullable. In JSON, this is equivalent "
            "to `type: [<type>, 'null']`"
        },
    )

    # Keywords for Any Instance Type
    # https://json-schema.org/draft/2020-12/json-schema-validation.html#name-validation-keywords-for-any
    type: object = field(
        default=None, metadata={"description": "The type annotation of the field."}
    )
    enum: list[T] | None = field(
        default=None,
        metadata={"description": "A list of allowed values.", "aliases": ["choices"]},
    )
    const: T = field(  # type: ignore
        default=Undefined,
        metadata={
            "description": "A single allowed value. functionally equivalent to an "
            "'enum' with a single value.",
        },
    )

    # Keywords for Numeric Instances (number and integer)
    # https://json-schema.org/draft/2020-12/json-schema-validation.html#name-validation-keywords-for-num
    minimum: float | None = field(
        default=None,
        metadata={
            "description": "The inclusive minimum allowed value.",
            "aliases": ["min", "ge"],
        },
    )
    maximum: float | None = field(
        default=None,
        metadata={
            "description": "The inclusive maximum allowed value.",
            "aliases": ["max", "le"],
        },
    )
    exclusive_minimum: float | None = field(
        default=None,
        metadata={
            "description": "The exclusive minimum allowed value.",
            "aliases": ["exclusiveMinimum", "gt"],
        },
    )
    exclusive_maximum: float | None = field(
        default=None,
        metadata={
            "description": "The exclusive maximum allowed value.",
            "aliases": ["exclusiveMaximum", "lt"],
        },
    )
    multiple_of: float | None = field(
        default=None,
        metadata={
            "description": "The allowed step size. Value is valid if (value / multiple_"
            "of) is an integer.",
            "aliases": ["multipleOf", "step"],
        },
    )
    # not in json schema, for Decimal types.  Also in pydantic.
    decimal_places: int | None = field(
        default=None,
        metadata={
            "descripion": "Maximum number of digits within the decimal. It does "
            "not include trailing decimal zeroes."
        },
    )

    # Keywords for String Instances
    # https://json-schema.org/draft/2020-12/json-schema-validation.html#name-validation-keywords-for-str
    min_length: int | None = field(
        default=None,
        metadata={
            "description": "The minimum allowed length. Must be >= 0.",
            "aliases": ["minLength"],
        },
    )
    max_length: int | None = field(
        default=None,
        metadata={
            "description": "The maximum allowed length. Must be >= 0.",
            "aliases": ["maxLength"],
        },
    )
    pattern: str | None = field(
        default=None,
        metadata={
            "description": "A regex pattern for the value.",
            "aliases": ["regex", "filter"],  # regex in pydantic, filter for FileEdit
        },
    )
    # NOTE: format is listed in this section, but needn't strictly apply to strings.
    format: JsonStringFormats | None = field(
        default=None, metadata={"description": "The format of the field."}
    )

    # Keywords for Sequence (Array) Instances
    # https://json-schema.org/draft/2020-12/json-schema-validation.html#name-validation-keywords-for-arr
    min_items: int | None = field(
        default=None,
        metadata={
            "description": "The (inclusive) minimum allowed number of items. "
            "Must be >= 0",
            # min_length/min_inclusive are from annotated_types
            "aliases": ["minItems", "min_length", "min_inclusive"],
        },
    )
    max_items: int | None = field(
        default=None,
        metadata={
            "description": "The (inclusive) maximum allowed number of items. "
            "Must be >= 0",
            "aliases": ["maxItems", "max_length"],  # max_length in annotated_types
        },
    )
    unique_items: bool | None = field(
        default=None,
        metadata={
            "description": "Whether the items in the list must be unique.",
            "aliases": ["uniqueItems"],
        },
    )

    # # Keywords for Mapping (Object) Instances
    # # https://json-schema.org/draft/2020-12/json-schema-validation.html#name-validation-keywords-for-obj  # noqa
    # max_properties: int | None = field(
    #     default=None,
    #     metadata=dict(description="The maximum allowed number of keys. Must be >= 0.")
    # )
    # min_properties: int | None = field(
    #     default=None,
    #     metadata=dict(description="The minimum allowed number of keys. Must be >= 0.")
    # )
    # required: list[str] | None = field(
    #     default=None,
    #     metadata=dict(
    #         description="A list of required keys that must be present in a mapping/object." # noqa
    #     ),
    # )

    read_only: bool | None = field(
        default=None,
        metadata={
            "description": "Whether the field is read-only. If True, the value of the "
            "instance is managed exclusively by the owning authority, and attempts by "
            "an application to modify the value of this property are expected to be "
            "ignored or rejected by that owning authority"
        },
    )

    # UI Specific
    widget: str | None = field(
        default=None,
        metadata={
            "description": "The name of the widget to use for this field. "
            "If not provided, the widget will be inferred from the type annotation."
        },
    )
    disabled: bool | None = field(
        default=None,
        metadata={
            "description": "Whether the widget should be disabled. Marking a field as "
            "read-only will render it greyed out, but its text value will be "
            "selectable. Disabling it will prevent its value to be selected at all."
        },
    )
    enum_disabled: list[T] | None = field(
        default=None,
        metadata={
            "description": "A list of values that should be disabled in a combobox "
            "widget."
        },
    )
    help: str | None = field(
        default=None,
        metadata={
            "description": "text next to a field to guide the end user filling it."
        },
    )
    placeholder: str | None = field(
        default=None,
        metadata={
            "description": "A placeholder string to display when the field is empty."
        },
    )
    visible: bool | None = field(
        default=None,
        metadata={
            "description": "Whether the field should be visible in the GUI. "
            "This is useful for hiding fields that are only used for validation."
        },
    )
    orientation: Literal["horizontal", "vertical"] | None = field(
        default=None,
        metadata={"description": "Orientation of the widget, for things like sliders."},
    )

    _native_field: Any | None = field(
        default=None,
        compare=False,
        hash=False,
        repr=False,
        metadata={
            "description": "Internal use only. If this field is derived from a native "
            "dataclasses.Field, or attrs.Attribute, or pydantic.fields.ModelField this "
            "will be a reference to that object."
        },
    )
    _original_annotation: Any | None = field(
        default=None,
        compare=False,
        hash=False,
        repr=False,
        metadata={
            "description": "Internal use only. If this field is derived from a "
            "typing.Annotated[...] annotation, this will be a reference to the origin "
            "annotation."
        },
    )

    def get_default(self) -> T | None:
        """Return the default value for this field."""
        return (
            self.default  # TODO: deepcopy mutable defaults?
            if self.default_factory is None
            else self.default_factory()
        )

    def asdict(self, include_unset: bool = True) -> dict[str, Any]:
        """Return the field as a dictionary.

        If `include_unset` is `False`, only fields that have been set will be included.
        """
        d = dc.asdict(self)
        if not include_unset:
            d = {
                k: v
                for k, v in d.items()
                if (v is not Undefined if k in ("default", "const") else v is not None)
            }
        return d

    def replace(self, **kwargs: Any) -> UiField:
        """Return a new Field with the given values replaced."""
        return dc.replace(self, **kwargs)

    @property
    def resolved_type(self) -> Any:
        """Return field type, resolving any forward references.

        Note that this will also return the origin type for Annotated types.
        """
        from magicgui._type_resolution import _try_cached_resolve

        return _try_cached_resolve(self.type)

    @property
    def is_annotated_type(self) -> bool:
        """Whether the field is an Annotated type."""
        return get_origin(self.type) is Annotated

    def parse_annotated(self) -> UiField[T]:
        """Extract info from Annotated type if present, and return new field.

        If self.type is not an Annotated type, return self.
        """
        if not self.is_annotated_type:
            return self

        kwargs = _uikwargs_from_annotated_type(self.type)

        if (
            self.default is not Undefined
            and kwargs.get("default", Undefined) is not Undefined
        ):
            warnings.warn(
                "Cannot set default value in both type annotation and field. Overriding"
                f" default {kwargs['default']} with {self.default} in field "
                f"{self.name!r}",
                stacklevel=2,
            )
            kwargs.pop("default", None)
        if self.name is not None and kwargs.get("name") is not None:
            warnings.warn(
                "Cannot set name in both type annotation and field. Overriding"
                f" name {kwargs['name']!r} with {self.name!r} in field {self.name!r}",
                stacklevel=2,
            )
            kwargs.pop("name", None)
        return dc.replace(self, **kwargs)

    def create_widget(self, value: T | _Undefined = Undefined) -> ValueWidget[T]:
        """Create a new Widget for this field."""
        from magicgui.type_map import get_widget_class

        # TODO: this should be cached in some way
        # Map uifield names to widget kwargs
        # FIXME: this part needs a lot of work.
        # This is the biggest challenge for integrating this new UiField idea
        # (which tries to map nicely to existing schemas like JSON Schema)
        # with the rest of the codebase, which used less "general" naming schemes.
        _name_map = {
            "name": "name",
            "visible": "visible",
            "nullable": "nullable",
            "orientation": "orientation",
            "type": "annotation",
            "enum": "choices",
            "title": "label",
            "description": "tooltip",
            "maximum": "max",
            "minimum": "min",
            # "title": "text",  # PushButton only
            # "exclusive_maximum": "stop",  # RangeEdit only
            # "minimum": "start", # RangeEdit only
            "multiple_of": "step",
            "widget": "widget_type",
        }

        d = (
            self.parse_annotated()
            .replace(_native_field=None)
            .asdict(include_unset=False)
        )
        opts = {_name_map[k]: v for k, v in d.items() if k in _name_map}
        if "disabled" in d:
            opts["enabled"] = not d["disabled"]

        # TODO: very hacky... but we don't have the concept of exclusive min/max
        # for float values.
        if "exclusive_maximum" in d:
            m = 1 if d.get("type") is int else 0.00000000001
            opts["max"] = d["exclusive_maximum"] - m
        if "exclusive_minimum" in d:
            m = 1 if d.get("type") is int else 0.00000000001
            opts["min"] = d["exclusive_minimum"] + m

        value = value if value is not Undefined else self.get_default()  # type: ignore
        cls, kwargs = get_widget_class(value=value, annotation=self.type, options=opts)
        return cls(**kwargs)  # type: ignore


_UI_FIELD_NAMES: set[str] = set()
_UI_FIELD_ALIASES: dict[str, str] = {}

# gather up all the aliases for the UiField fields
# so we can use them in _rename_aliases
for field_info in dc.fields(UiField):
    _UI_FIELD_NAMES.add(field_info.name)
    for alias in field_info.metadata.get("aliases", []):
        _UI_FIELD_ALIASES[alias] = field_info.name


def _rename_aliases(input: dict[str, Any]) -> dict[str, Any]:
    """Rename any aliases in the input dict to their accepted names."""
    return {_UI_FIELD_ALIASES.get(k, k): v for k, v in input.items()}


def _uikwargs_from_annotated_type(hint: Any) -> dict[str, Any]:
    """Convert an Annotated type to a dict of UiField kwargs."""
    # hint must be an Annotated[...] type

    annotated_types = sys.modules.get("annotated_types")
    base_metas: list[BaseMetadata] = []

    origin, *metadata = get_args(hint)
    kwargs = {}
    for item in metadata:
        if isinstance(item, UiField):
            kwargs.update(item.asdict(include_unset=False))
        elif annotated_types is not None:
            # annotated_types >= 0.3.0 is supported
            if isinstance(item, annotated_types.BaseMetadata):
                base_metas.append(item)
            elif isinstance(item, annotated_types.GroupedMetadata):
                base_metas.extend(item)
        # TODO: support pydantic.fields.FieldInfo?
        # TODO: support re.Pattern?

    if base_metas:
        _annotated_kwargs = {}
        for i in base_metas:
            _annotated_kwargs.update(dc.asdict(i))
        if "max_exclusive" in _annotated_kwargs:
            _annotated_kwargs["max_items"] = _annotated_kwargs.pop("max_exclusive") - 1
        kwargs.update(_rename_aliases(_annotated_kwargs))

    kwargs.update({"type": origin, "_original_annotation": hint})
    return kwargs


def _uifield_from_dataclass(field: dc.Field) -> UiField:
    """Create a UiField from a dataclass field."""
    default = field.default if field.default is not dc.MISSING else Undefined
    dfactory = (
        field.default_factory if field.default_factory is not dc.MISSING else None
    )
    extra = {k: v for k, v in field.metadata.items() if k in _UI_FIELD_NAMES}

    return UiField(
        name=field.name,
        type=field.type,
        default=default,
        default_factory=dfactory,
        _native_field=field,
        **extra,
    )


def _uifield_from_attrs(field: Attribute) -> UiField:
    """Create a UiField from an attrs field."""
    from attrs import NOTHING, Factory

    default = field.default if field.default is not NOTHING else Undefined
    default_factory = None
    if isinstance(default, Factory):
        default_factory = default.factory
        default = Undefined

    extra = {k: v for k, v in field.metadata.items() if k in _UI_FIELD_NAMES}

    return UiField(
        name=field.name,
        type=field.type,
        default=default,
        default_factory=default_factory,
        _native_field=field,
        **extra,
    )


def _uifield_from_pydantic1(model_field: ModelField) -> UiField:
    """Create a UiField from a pydantic ModelField."""
    from pydantic.fields import SHAPE_SINGLETON
    from pydantic.fields import Undefined as PydanticUndefined

    finfo = model_field.field_info

    _extra_dict = finfo.extra.copy()
    # backport from pydantic2
    if "json_schema_extra" in _extra_dict:
        _extra_dict.update(_extra_dict.pop("json_schema_extra"))

    extra = {k: v for k, v in _extra_dict.items() if k in _UI_FIELD_NAMES}
    const = finfo.const if finfo.const not in (None, PydanticUndefined) else Undefined
    default = (
        Undefined if finfo.default in (PydanticUndefined, Ellipsis) else finfo.default
    )

    nullable = None
    if model_field.allow_none and (
        model_field.shape != SHAPE_SINGLETON or not model_field.sub_fields
    ):
        nullable = True

    return UiField(
        name=model_field.name,
        title=finfo.title,
        description=finfo.description,
        default=default,
        default_factory=model_field.default_factory,
        type=model_field.outer_type_,
        nullable=nullable,
        const=const,
        minimum=finfo.ge,
        maximum=finfo.le,
        exclusive_minimum=finfo.gt,
        exclusive_maximum=finfo.lt,
        multiple_of=finfo.multiple_of,
        min_length=finfo.min_length,
        max_length=finfo.max_length,
        pattern=finfo.regex,
        # format=finfo.format,
        min_items=finfo.min_items,
        max_items=finfo.max_items,
        unique_items=finfo.unique_items,
        _native_field=model_field,
        **extra,
    )


def _uifield_from_pydantic2(finfo: FieldInfo, name: str) -> UiField:
    """Create a UiField from a pydantic ModelField."""
    import annotated_types as at
    from pydantic_core import PydanticUndefined

    if isinstance(finfo.json_schema_extra, dict):
        extra = {
            k: v for k, v in finfo.json_schema_extra.items() if k in _UI_FIELD_NAMES
        }
    else:
        extra = {}
    default = (
        Undefined if finfo.default in (PydanticUndefined, Ellipsis) else finfo.default
    )

    nullable = None
    if get_origin(finfo.annotation) is Union and any(
        i for i in get_args(finfo.annotation) if i is type(None)
    ):
        nullable = True

    restrictions: dict = {}
    for meta in finfo.metadata:
        if isinstance(meta, at.Ge):
            restrictions["minimum"] = meta.ge
        elif isinstance(meta, at.Gt):
            restrictions["exclusive_minimum"] = meta.gt
        elif isinstance(meta, at.Le):
            restrictions["maximum"] = meta.le
        elif isinstance(meta, at.Lt):
            restrictions["exclusive_maximum"] = meta.lt
        elif isinstance(meta, at.MultipleOf):
            restrictions["multiple_of"] = meta.multiple_of
        elif isinstance(meta, at.MinLen):
            restrictions["min_length"] = meta.min_length
        elif isinstance(meta, at.MaxLen):
            restrictions["max_length"] = meta.max_length
        elif hasattr(meta, "__dict__"):
            # PydanticGeneralMetadata
            restrictions["pattern"] = meta.__dict__.get("pattern")

    return UiField(
        name=name,
        title=finfo.title,
        description=finfo.description,
        default=default,
        default_factory=finfo.default_factory,
        type=finfo.annotation,
        nullable=nullable,
        # const=const,
        **restrictions,
        # format=finfo.format,
        _native_field=finfo,
        **extra,
    )


# TODO:
class _ContainerFields:
    autofocus: str | None = field(
        default=None,
        metadata={"description": "Name of a field that should be autofocused on."},
    )
    labels: list[str] | bool | None = field(
        default=None,
        metadata={
            "description": "If True, all fields will be labeled. If False, no fields "
            "will be labeled. If a list, only the fields in the list will be labeled."
        },
    )


def _is_attrs_model(obj: Any) -> TypeGuard[HasAttrs]:
    return getattr(obj, "__attrs_attrs__", None) is not None


def _get_pydantic_model(cls: type) -> type[pydantic.BaseModel] | None:
    pydantic = sys.modules.get("pydantic")
    if pydantic is not None:
        if isinstance(cls, type) and issubclass(cls, pydantic.BaseModel):
            return cls
        elif isinstance(cls, pydantic.BaseModel):
            return type(cls)
        elif hasattr(cls, "__pydantic_model__"):
            return _get_pydantic_model(cls.__pydantic_model__)
    return None


def _get_function_defaults(func: FunctionType) -> dict[str, Any]:
    """Return a dict of the default values for a function's parameters."""
    # extracted bit from inspect.signature... ~20x faster
    pos_count = func.__code__.co_argcount
    arg_names = func.__code__.co_varnames

    defaults = func.__defaults__ or ()

    non_default_count = pos_count - len(defaults)
    positional_args = arg_names[:pos_count]

    output = {
        name: defaults[offset]
        for offset, name in enumerate(positional_args[non_default_count:])
    }
    if func.__kwdefaults__:
        output.update(func.__kwdefaults__)
    return output


def _ui_fields_from_annotation(cls: type) -> Iterator[UiField]:
    """Iterate UiFields extracted from object __annotations__."""
    # fallback for typed dict, named tuples, & functions

    annotations: dict = getattr(cls, "__annotations__", {})
    if not annotations and isinstance(annotations, dict):  # pragma: no cover
        raise TypeError(
            f"Cannot create  from object {type(cls)} without `__annotations__`"
        )

    # named tuples have _fields and _field_defaults
    field_names = cls._fields if hasattr(cls, "_fields") else annotations
    if isinstance(cls, FunctionType):
        defaults = _get_function_defaults(cls)
    else:
        defaults = getattr(cls, "_field_defaults", {})

    for name in field_names:
        field = UiField(
            name=name,
            type=annotations.get(name),
            default=defaults.get(name, Undefined),
        )
        yield field.parse_annotated()


def _iter_ui_fields(object: Any) -> Iterator[UiField]:
    # check if it's a pydantic model
    model = _get_pydantic_model(object)
    if model is not None:
        if hasattr(model, "model_fields"):
            for name, field_info in model.model_fields.items():
                yield _uifield_from_pydantic2(field_info, name)
        else:
            for pf in model.__fields__.values():
                yield _uifield_from_pydantic1(pf)
        return

    if hasattr(object, "__pydantic_fields__"):
        # pydantic2 style dataclass
        for name, field_info in object.__pydantic_fields__.items():
            yield _uifield_from_pydantic2(field_info, name)
        return

    # check if it's a (non-pydantic) dataclass
    if dc.is_dataclass(object):
        for df in dc.fields(object):
            yield _uifield_from_dataclass(df)
        return

    # check if it's an attrs class
    if _is_attrs_model(object):
        for af in object.__attrs_attrs__:
            yield _uifield_from_attrs(af)
        return

    # fallback to looking at __annotations__ (named tuple, typed dict, function)
    if hasattr(object, "__annotations__"):
        yield from _ui_fields_from_annotation(object)
        return

    raise TypeError(
        f"{object} is not a dataclass, attrs, or pydantic, model"
    )  # pragma: no cover


@lru_cache(maxsize=None)
def _cached_iter_ui_fields(cls: type) -> tuple[UiField, ...]:
    return tuple(_iter_ui_fields(cls))


def get_ui_fields(cls_or_instance: object) -> tuple[UiField, ...]:
    """Derive UIFields from an object.

    Parameters
    ----------
    cls_or_instance : Any
        Object to extract fields from.  Could be a dataclass or dataclass instance,
        an attrs class or instance, a pydantic model or instance, a named tuple,
        typed dict, or a function.
    """
    try:
        if isinstance(cls_or_instance, (type, FunctionType)):
            return _cached_iter_ui_fields(cls_or_instance)
        return _cached_iter_ui_fields(type(cls_or_instance))  # type: ignore
    except TypeError:
        return tuple(_iter_ui_fields(cls_or_instance))


def _uifields_to_container(
    ui_fields: Iterable[UiField],
    values: Mapping[str, Any] | None = None,
    *,
    container_kwargs: Mapping | None = None,
) -> ContainerWidget[ValueWidget]:
    """Create a container widget from a sequence of UiFields.

    This function is the heart of build_widget.

    Parameters
    ----------
    ui_fields : Iterable[UiField]
        A sequence of UiFields to use to create the container.
    values : Mapping[str, Any], optional
        A mapping of field name to values to use to initialize each widget the
        container, by default None.
    container_kwargs : Mapping, optional
        A mapping of keyword arguments to pass to the container constructor,
        by default None.

    Returns
    -------
    ContainerWidget[ValueWidget]
        A container widget with a widget for each UiField.
    """
    from magicgui import widgets

    container = widgets.Container(
        widgets=[field.create_widget() for field in ui_fields],
        **(container_kwargs or {}),
    )
    if values is not None:
        container.update(values)
    return container


def _get_values(obj: Any) -> dict | None:
    """Return a dict of values from an object.

    The object can be a dataclass, attrs, pydantic object or named tuple.
    """
    if isinstance(obj, dict):
        return obj

    # named tuple
    if isinstance(obj, tuple) and hasattr(obj, "_asdict"):
        return cast(dict, obj._asdict())

    # dataclass
    if dc.is_dataclass(type(obj)):
        return dc.asdict(obj)

    # attrs
    attr = sys.modules.get("attr")
    if attr is not None and attr.has(obj):
        return cast(dict, attr.asdict(obj))

    # pydantic models
    if hasattr(obj, "model_dump"):
        return cast(dict, obj.model_dump())
    elif hasattr(obj, "dict"):
        return cast(dict, obj.dict())

    return None


# TODO: unify this with magicgui
def build_widget(cls_or_instance: Any) -> ContainerWidget[ValueWidget]:
    """Build a magicgui widget from a dataclass, attrs, pydantic, or function."""
    values = None if isinstance(cls_or_instance, type) else _get_values(cls_or_instance)
    return _uifields_to_container(get_ui_fields(cls_or_instance), values=values)



#!/usr/bin/env python

"""Tests return widget types"""

import pathlib
from datetime import date, datetime, time
from inspect import signature

import numpy as np
import pytest

import magicgui
from magicgui import widgets


def _dataframe_equals(object1, object2):
    assert object1.equals(object2)


def _ndarray_equals(object1: np.ndarray, object2: np.ndarray):
    assert np.array_equal(object1, object2)


def _default_equals(object1, object2):
    assert object1 == object2


def _generate_pandas_test_data():
    pd = pytest.importorskip(
        "pandas", reason="Pandas required for a return widget test"
    )
    return pd.DataFrame({"Res1": [1, 2, 3], "Res2": [4, 5, 6]})


parameterizations = [
    # pandas dataframe
    (
        _generate_pandas_test_data(),
        widgets.Table,
        _dataframe_equals,
    ),
    # numpy array
    (
        np.array([1, 1, 1, 2, 2, 2, 3, 3, 3]).reshape((3, 3)),
        widgets.Table,
        _ndarray_equals,
    ),
    # NOTE: disabling for now... these types are too broad to choose a table
    # # dict
    # ({"a": [1], "b": [2], "c": [3]}, widgets.Table, _default_equals),
    # # list
    # ([1, 2, 3], widgets.Table, _default_equals),
    # # tuple
    # (([[1], [2], [3]], ["A", "B", "C"], ["Row 1"]), widgets.Table, _default_equals),
    # boolean
    (True, widgets.LineEdit, _default_equals),
    # int
    (5, widgets.LineEdit, _default_equals),
    # float
    (3.5, widgets.LineEdit, _default_equals),
    # string
    ("foo", widgets.LineEdit, _default_equals),
    # path
    (pathlib.Path().absolute(), widgets.LineEdit, _default_equals),
    # date
    (date.today(), widgets.LineEdit, _default_equals),
    # time
    (time(), widgets.LineEdit, _default_equals),
    # datetime
    (datetime.now(), widgets.LineEdit, _default_equals),
    # range
    (range(10), widgets.LineEdit, _default_equals),
    # slice
    (slice(6), widgets.LineEdit, _default_equals),
]


def generate_magicgui(data):
    def func():
        return data

    sig = signature(func)
    func.__signature__ = sig.replace(return_annotation=type(data))

    return magicgui.magicgui(func, result_widget=True)


@pytest.mark.parametrize("data, expected_type, equality_check", parameterizations)
def test_return_widget_for_type(data, expected_type, equality_check):
    widget = generate_magicgui(data)
    equality_check(widget(), data)
    assert isinstance(widget._result_widget, expected_type)


def test_table_return_annotation():
    @magicgui.magicgui(result_widget=True)
    def f() -> "magicgui.widgets.Table": ...

    @magicgui.magicgui(result_widget=True)
    def f2() -> widgets.Table: ...

    assert isinstance(f._result_widget, widgets.Table)
    assert isinstance(f2._result_widget, widgets.Table)



"""# Basic example

A basic example using magicgui.
"""

from magicgui import magicgui


@magicgui
def example(x: int, y="hi"):
    """Basic example function."""
    return x, y


example.changed.connect(print)
example.show(run=True)



from magicgui.widgets import protocols


class _OrientationMixin:
    """Properties for classes wrapping widgets that support orientation."""

    _widget: protocols.SupportsOrientation

    @property
    def orientation(self) -> str:
        """Orientation of the widget."""
        return self._widget._mgui_get_orientation()

    @orientation.setter
    def orientation(self, value: str) -> None:
        if value not in {"horizontal", "vertical"}:
            raise ValueError(
                "Only horizontal and vertical orientation are currently supported"
            )
        self._widget._mgui_set_orientation(value)


class _ReadOnlyMixin:
    """Properties for classes wrapping widgets that support read-only."""

    _widget: protocols.SupportsReadOnly

    @property
    def read_only(self) -> bool:
        """read_only of the widget."""
        return self._widget._mgui_get_read_only()

    @read_only.setter
    def read_only(self, value: bool) -> None:
        self._widget._mgui_set_read_only(value)



from __future__ import annotations

from enum import Enum, EnumMeta
from typing import TYPE_CHECKING, Any, Callable, cast

from magicgui.types import ChoicesType, Undefined, _Undefined

from ._value_widget import T, ValueWidget

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from magicgui.widgets import protocols

    from ._widget import WidgetKwargs


class CategoricalWidget(ValueWidget[T]):
    """Widget with a value and choices.  Wraps CategoricalWidgetProtocol.

    Parameters
    ----------
    value : Any, optional
        The initially selected choice.
    choices : Enum, Iterable, or Callable
        Available choices displayed in the combo box.
    bind : Callable[[ValueWidget], Any] | Any, optional
        A value or callback to bind this widget. If provided, whenever
        [`widget.value`][magicgui.widgets.bases.ValueWidget.value] is
        accessed, the value provided here will be returned instead. `bind` may be a
        callable, in which case `bind(self)` will be returned (i.e. your bound callback
        must accept a single parameter, which is this widget instance).
    nullable : bool, optional
        If `True`, the widget will accepts `None` as a valid value, by default `False`.
    **base_widget_kwargs : Any
        All additional keyword arguments are passed to the base
        [`magicgui.widgets.Widget`][magicgui.widgets.Widget] constructor.
    """

    _widget: protocols.CategoricalWidgetProtocol
    null_string: str = "-----"
    _allow_multiple = False

    def __init__(
        self,
        value: T | _Undefined = Undefined,
        choices: ChoicesType = (),
        *,
        allow_multiple: bool | None = None,
        bind: T | Callable[[ValueWidget], T] | _Undefined = Undefined,
        nullable: bool = False,
        **base_widget_kwargs: Unpack[WidgetKwargs],
    ) -> None:
        if allow_multiple is not None:
            self._allow_multiple = allow_multiple
        self._default_choices = choices
        super().__init__(
            value=value, bind=bind, nullable=nullable, **base_widget_kwargs
        )

    def _post_init(self) -> None:
        super()._post_init()
        self.reset_choices()
        self.native_parent_changed.connect(self.reset_choices)

    @property
    def value(self) -> T:
        """Return current value of the widget."""
        return ValueWidget.value.fget(self)  # type: ignore

    @value.setter
    def value(self, value: T) -> None:
        if isinstance(value, (list, tuple)) and self._allow_multiple:
            if any(v not in self.choices for v in value):
                raise ValueError(
                    f"{value!r} is not a valid choice. must be in {self.choices}"
                )
        elif value not in self.choices:
            raise ValueError(
                f"{value!r} is not a valid choice. must be in {self.choices}"
            )
        return ValueWidget.value.fset(self, value)  # type: ignore

    @property
    def options(self) -> dict:
        """Return options currently being used in this widget."""
        d = super().options.copy()
        d.update({"choices": self._default_choices})
        return d

    def reset_choices(self, *_: Any) -> None:
        """Reset choices to the default state.

        If self._default_choices is a callable, this may NOT be the exact same set of
        choices as when the widget was instantiated, if the callable relies on external
        state.
        """
        self.choices = self._default_choices  # type: ignore

    @property
    def current_choice(self) -> str:
        """Return the text of the currently selected choice."""
        return self._widget._mgui_get_current_choice()

    def __len__(self) -> int:
        """Return the number of choices."""
        return self._widget._mgui_get_count()

    def get_choice(self, choice_name: str) -> T:
        """Get data for the provided ``choice_name``."""
        return cast(T, self._widget._mgui_get_choice(choice_name))

    def set_choice(self, choice_name: str, data: Any | None = None) -> None:
        """Set data for the provided ``choice_name``."""
        data = data if data is not None else choice_name
        self._widget._mgui_set_choice(choice_name, data)
        if choice_name == self.current_choice:
            self.changed.emit(self.value)

    def del_choice(self, choice_name: str) -> None:
        """Delete the provided ``choice_name`` and associated data."""
        self._widget._mgui_del_choice(choice_name)

    @property
    def choices(self) -> tuple[T | None, ...]:
        """Available value choices for this widget."""
        _choices = tuple(i[1] for i in self._widget._mgui_get_choices())
        if self._nullable and None not in _choices:
            return (*_choices, None)
        return _choices

    @choices.setter
    def choices(self, choices: ChoicesType) -> None:
        str_func: Callable = _get_name if isinstance(choices, EnumMeta) else str  # type: ignore
        if isinstance(choices, dict):
            if "choices" not in choices or "key" not in choices:
                raise ValueError(
                    "When setting choices with a dict, the dict must have keys "
                    "'choices' (Iterable), and 'key' (callable that takes a each value "
                    "in `choices` and returns a string."
                )
            _choices = choices["choices"]
            str_func = choices["key"]
        elif not isinstance(choices, EnumMeta) and callable(choices):
            _choices = choices(self)
            # ensure that when setting choices with a callable,
            # the default choices are also updated with that callable
            self._default_choices = choices
        else:
            _choices = choices

        _normed: list[tuple[str, Any]] = list(_choices)
        if not all(isinstance(i, tuple) and len(i) == 2 for i in _normed):
            _normed = [(str_func(i), i) for i in _choices]
        if self._nullable:
            _normed.insert(0, (self.null_string, self.null_value))
        return self._widget._mgui_set_choices(_normed)


def _get_name(obj: Enum) -> str:
    return str(obj.name)



"""# Chaining functions together

This example demonstrates chaining multiple functions together.
"""

from magicgui import magicgui, widgets


@magicgui(auto_call=True)
def func_a(x: int = 64, y: int = 64):
    """Callable function A."""
    print("calling func_a")
    return x + y


@magicgui(auto_call=True, input={"visible": False, "label": " ", "max": 100000})
def func_b(input: int, mult=1.0):
    """Callable function B."""
    print("calling func_b")
    result = input * mult
    # since these function defs live in globals(), you can update them directly
    func_c.input.value = result
    return result


# alternatively, you can the `widget.called` signal to connect a callback function
# where the result of the function being called is at `value`
@func_a.called.connect
def _on_func_a(value: str):
    func_b.input.value = value


@magicgui(
    auto_call=True,
    input={"visible": False, "max": 100000},
    result_widget=True,
    labels=False,
)
def func_c(input: int, format: str = "({} + {}) * {} is {}") -> str:
    """Callable function C."""
    print("calling func_c\n")
    return format.format(func_a.x.value, func_a.y.value, func_b.mult.value, input)


container = widgets.Container(
    widgets=[func_a, func_b, func_c], layout="vertical", labels=False
)
container.native.setMinimumWidth(500)
func_a()
container.show(run=True)

# notice which functions get called when you change each widget.



"""Widget base classes.

These will rarely be used directly by end-users, instead see the "concrete" widgets
exported in :mod:`magicgui.widgets`.

All magicgui [Widget][magicgui.widgets.Widget] bases comprise a backend widget that
implements one of the widget protocols defined in
[`magicgui.widgets.protocols`](protocols). The basic composition pattern is as follows:

.. code-block:: python

   class Widget:

       def __init__(
            self,

            # widget_type is a class, likely from the `backends` module
            # that implements one of the `WidgetProtocols` defined in _protocols.
            widget_type: Type[protocols.WidgetProtocol],

            # backend_kwargs is a key-value map of arguments that will be provided
            # to the concrete (backend) implementation of the WidgetProtocol
            backend_kwargs: dict = dict(),

            # additional kwargs will be provided to the magicgui.Widget itself
            # things like, `name`, `value`, etc...
            **kwargs
        ):
           # instantiation of the backend widget.
           self._widget = widget_type(**backend_kwargs)

           # ... go on to set other kwargs


These widgets are unlikely to be instantiated directly, (unless you're creating a custom
widget that implements one of the WidgetProtocols... as the backed widgets do).
Instead, one will usually instantiate the widgets in :mod:`magicgui.widgets._concrete`
which are all available directly in the :mod:`magicgui.widgets` namespace.

The :func:`~magicgui.widgets.create_widget` factory function may be used to
create a widget subclass appropriate for the arguments passed (such as "value" or
"annotation").

"""

from ._button_widget import ButtonWidget
from ._categorical_widget import CategoricalWidget
from ._container_widget import ContainerWidget, DialogWidget, MainWindowWidget
from ._create_widget import create_widget
from ._ranged_widget import RangedWidget, TransformedRangedWidget
from ._slider_widget import MultiValuedSliderWidget, SliderWidget
from ._toolbar import ToolBarWidget
from ._value_widget import ValueWidget
from ._widget import Widget

__all__ = [
    "ButtonWidget",
    "CategoricalWidget",
    "ContainerWidget",
    "create_widget",
    "DialogWidget",
    "MainWindowWidget",
    "MultiValuedSliderWidget",
    "RangedWidget",
    "SliderWidget",
    "ToolBarWidget",
    "TransformedRangedWidget",
    "ValueWidget",
    "Widget",
]



"""# napari Qt demo

Napari provides a few conveniences with magicgui, and one of the
most commonly used is the layer combo box that gets created when
a parameter is annotated as napari.layers.Layer.

The layer box will stay in sync with the viewer model, adding and
removing layers as needed.

This example shows how to use just that widget in the context
of a larger custom QWidget.
"""

import napari
from qtpy.QtWidgets import QVBoxLayout, QWidget

from magicgui.widgets import create_widget


class CustomWidget(QWidget):
    """A custom widget class."""

    def __init__(self) -> None:
        super().__init__()
        self.setLayout(QVBoxLayout())
        # change annotation to napari.layers.Image (e.g) to restrict to just Images
        self._layer_combo = create_widget(annotation=napari.layers.Layer)
        # magicgui widgets hold the Qt widget at `widget.native`
        self.layout().addWidget(self._layer_combo.native)


viewer = napari.Viewer()
viewer.add_points()
viewer.add_points()

my_widget = CustomWidget()
viewer.window.add_dock_widget(my_widget)

# when my_widget is a magicgui.Widget, it will detect that it has been added
# to a viewer, and automatically update the choices.  Otherwise, you need to
# trigger this yourself:
my_widget._layer_combo.reset_choices()
viewer.layers.events.inserted.connect(my_widget._layer_combo.reset_choices)
viewer.layers.events.removed.connect(my_widget._layer_combo.reset_choices)

napari.run()



"""# Self reference magicgui widgets

Widgets created with magicgui can reference themselves, and use the widget API.
"""

from magicgui import magicgui


@magicgui(auto_call=True, width={"max": 800, "min": 100}, x={"widget_type": "Slider"})
def function(width=400, x: int = 50):
    """Example function."""
    # the widget can reference itself, and use the widget API
    function.x.width = width


function.show(run=True)



import pytest
from typing_extensions import Annotated

from magicgui.signature import magic_signature, make_annotated


def test_make_annotated_raises():
    """Test options to annotated must be a dict."""
    with pytest.raises(TypeError):
        make_annotated(int, "not a dict")  # type: ignore

    with pytest.raises(TypeError):
        make_annotated(Annotated[int, "string"])

    make_annotated(Annotated[int, (("a", 1),)])  # this is ok


def test_make_annotated_works_with_already_annotated():
    """Test that make_annotated merges options with Annotated types."""
    annotated_type = Annotated[int, {"max": 10}]
    assert make_annotated(annotated_type) == annotated_type
    assert (
        make_annotated(annotated_type, {"min": 1})
        == Annotated[int, {"max": 10, "min": 1}]
    )


def _sample_func(a: int, b: str = "hi"):
    pass


def test_magic_signature_raises():
    """Test that gui_options must have keys that are params in function."""
    with pytest.raises(ValueError):
        magic_signature(_sample_func, gui_options={"not_a_param": {"choices": []}})


def test_signature_to_container():
    """Test that a MagicSignature can make a container."""
    sig = magic_signature(_sample_func, gui_options={"a": {"widget_type": "Slider"}})
    container = sig.to_container()
    assert len(container) == 2
    assert repr(container) == "<Container (a: int = 0, b: str = 'hi')>"
    assert repr(container.a) == "Slider(value=0, annotation=<class 'int'>, name='a')"

    param_repr = """<MagicParameter "a: int" {'widget_type': 'Slider'}>"""
    assert repr(sig.parameters["a"]) == param_repr



"""Magicgui application object, wrapping a backend application."""

from __future__ import annotations

import signal
from contextlib import contextmanager
from importlib import import_module
from typing import TYPE_CHECKING, Any, Callable, Iterator, Union

from magicgui.backends import BACKENDS

if TYPE_CHECKING:
    from types import ModuleType

    from magicgui.widgets.protocols import BaseApplicationBackend
DEFAULT_BACKEND = "qt"
APPLICATION_NAME = "magicgui"


@contextmanager
def event_loop(backend: str | None = None) -> Iterator[Application]:
    """Start an event loop in which to run the application."""
    with Application(backend) as app:
        try:
            yield app
        except Exception as e:
            print("An error was encountered in the event loop:\n", e)


class Application:
    """Magicgui Application, wrapping a native BaseApplicationBackend implementation."""

    _backend_module: ModuleType
    _backend: BaseApplicationBackend
    _instance: Application | None = None

    def __init__(self, backend_name: str | None = None) -> None:
        self._use(backend_name)

    @property
    def backend_name(self) -> str:
        """Return name of the GUI backend that this app wraps."""
        return "" if self._backend is None else self._backend._mgui_get_backend_name()

    @property
    def backend_module(self) -> ModuleType:
        """Return module object that defines the backend."""
        return self._backend_module

    def _use(self, backend_name: str | None = None) -> None:
        """Select a backend by name."""
        if not backend_name:
            backend_name = DEFAULT_BACKEND
        if not backend_name or backend_name.lower() not in BACKENDS:
            raise ValueError(
                f"backend_name must be one of {set(BACKENDS)!r}, "
                f"not {backend_name!r}"
            )

        module_name, _ = BACKENDS[backend_name]
        self._backend_module = import_module(f"magicgui.backends.{module_name}")
        self._backend = self.get_obj("ApplicationBackend")()

    def get_obj(self, name: str) -> Any:
        """Get the backend object for the given ``name`` (such as a widget)."""
        try:
            return getattr(self.backend_module, name)
        except AttributeError as e:
            raise AttributeError(
                f"Could not import object {name!r} from backend {self.backend_module}"
            ) from e

    def run(self) -> None:
        """Enter the native GUI event loop."""
        return self._backend._mgui_run()

    @property
    def native(self) -> Any:
        """Return the native GUI application instance."""
        return self._backend._mgui_get_native_app()

    def quit(self) -> None:
        """Quit the native GUI event loop."""
        return self._backend._mgui_quit()

    def create(self) -> None:
        """Create the native application."""
        # Ensure that the native app exists
        self.native  # noqa

    def process_events(self) -> None:
        """Process all pending GUI events."""
        return self._backend._mgui_process_events()

    def __repr__(self) -> str:
        """Return repr for this instance."""
        if not self.backend_name:
            return "<magicgui app with no backend>"
        else:
            return f"<magicgui app, wrapping the {self.backend_name} GUI toolkit>"

    def __enter__(self) -> Application:
        """Context manager to start this application."""
        self.create()
        return self

    def __exit__(self, *exc_details: Any) -> None:
        """Exit context manager for this application."""
        # enable ctrl-C
        signal.signal(signal.SIGINT, lambda *a: self.quit())
        self._backend._mgui_start_timer(500, lambda: None)
        self._backend._mgui_run()
        self._backend._mgui_stop_timer()

    def start_timer(
        self,
        interval: int = 1000,
        on_timeout: Callable[[], None] | None = None,
        single_shot: bool = False,
    ) -> None:
        """Start a timer with a given interval, optional callback, and single_shot."""
        self._backend._mgui_start_timer(interval, on_timeout, single=single_shot)


def _use_app(backend_name: str | None = None) -> Application:
    """Get/create the default Application object.

    It is safe to call this function multiple times, as long as
    backend_name is None or matches the already selected backend.

    Parameters
    ----------
    backend_name : str | None
        The name of the backend application to use. If not specified, Vispy
        tries to select a backend automatically. See ``vispy.use()`` for
        details.
    """
    # If we already have a default_app, raise error or return
    current = Application._instance
    if current is not None:
        if backend_name:
            names = current.backend_name.lower().replace("(", " ").strip(") ")
            _nm = [n for n in names.split(" ") if n]
            if backend_name.lower() not in _nm:
                raise RuntimeError(
                    f"Can only select a backend once, already using {_nm}."
                )
        else:
            return current  # Current backend matches backend_name

    # Create default app
    Application._instance = Application(backend_name)
    return Application._instance


AppRef = Union[Application, str, None]


def use_app(app: AppRef | None = None) -> Application:
    """Get/create the default Application object.  See _use_app docstring."""
    if app is None:
        return _use_app()
    elif isinstance(app, Application):
        return app
    elif isinstance(app, str):
        Application._instance = Application(app)
        return Application._instance
    raise TypeError(f"'app' must be string, Application, or None, got: {app!r}.  ")



"""# Snell's law demonstration using magicgui

Demo app for calculating angles of refraction according to Snell's law.
"""

import math
from enum import Enum

from magicgui import magicgui


class Medium(Enum):
    """Enum for various media and their refractive indices."""

    Glass = 1.520
    Oil = 1.515
    Water = 1.333
    Air = 1.0003


@magicgui(call_button="calculate", result_widget=True)
def snells_law(aoi=1.0, n1=Medium.Glass, n2=Medium.Water, degrees=True):
    """Calculate the angle of refraction given two media and an AOI."""
    if degrees:
        aoi = math.radians(aoi)
    try:
        n1 = n1.value
        n2 = n2.value
        result = math.asin(n1 * math.sin(aoi) / n2)
        return round(math.degrees(result) if degrees else result, 2)
    except ValueError:  # math domain error
        return "TIR!"


snells_law.show(run=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Iterable, get_type_hints

try:
    import ipywidgets
    from ipywidgets import widgets as ipywdg
except ImportError as e:
    raise ImportError(
        "magicgui requires ipywidgets to be installed to use the 'ipynb' backend. "
        "Please run `pip install ipywidgets`"
    ) from e


from magicgui.types import Separator
from magicgui.widgets import protocols

if TYPE_CHECKING:
    from magicgui.widgets.bases import Widget


def _pxstr2int(pxstr: int | str) -> int:
    if isinstance(pxstr, int):
        return pxstr
    if isinstance(pxstr, str) and pxstr.endswith("px"):
        return int(pxstr[:-2])
    return int(pxstr)


def _int2pxstr(pxint: int | str) -> str:
    return f"{pxint}px" if isinstance(pxint, int) else pxint


class _IPyWidget(protocols.WidgetProtocol):
    _ipywidget: ipywdg.Widget

    def __init__(
        self,
        wdg_class: type[ipywdg.Widget] | None = None,
        parent: ipywdg.Widget | None = None,
    ):
        if wdg_class is None:
            wdg_class = get_type_hints(self, None, globals()).get("_ipywidget")
        if wdg_class is None:
            raise TypeError("Must provide a valid ipywidget type")
        self._ipywidget = wdg_class()
        # TODO: assign parent

    def _mgui_close_widget(self):
        self._ipywidget.close()

    # `layout.display` will hide and unhide the widget and collapse the space
    # `layout.visibility` will make the widget (in)visible without changing layout
    def _mgui_get_visible(self):
        return self._ipywidget.layout.display != "none"

    def _mgui_set_visible(self, value: bool):
        self._ipywidget.layout.display = "block" if value else "none"

    def _mgui_get_enabled(self) -> bool:
        return not self._ipywidget.disabled

    def _mgui_set_enabled(self, enabled: bool):
        self._ipywidget.disabled = not enabled

    def _mgui_get_parent(self):
        # TODO: how does ipywidgets handle this?
        # return getattr(self._ipywidget, "parent", None)
        raise NotImplementedError(
            "parent not implemented for ipywidgets backend.  Please open an issue"
        )

    def _mgui_set_parent(self, widget):
        # TODO: how does ipywidgets handle this?
        self._ipywidget.parent = widget

    def _mgui_get_native_widget(self) -> ipywdg.Widget:
        return self._ipywidget

    def _mgui_get_root_native_widget(self) -> ipywdg.Widget:
        return self._ipywidget

    def _mgui_get_width(self) -> int:
        # TODO: ipywidgets deals in CSS ... by default width is `None`
        # will this always work with our base Widget assumptions?
        return _pxstr2int(self._ipywidget.layout.width)

    def _mgui_set_width(self, value: int | str) -> None:
        """Set the current width of the widget."""
        self._ipywidget.layout.width = _int2pxstr(value)

    def _mgui_get_min_width(self) -> int:
        return _pxstr2int(self._ipywidget.layout.min_width)

    def _mgui_set_min_width(self, value: int | str):
        self._ipywidget.layout.min_width = _int2pxstr(value)

    def _mgui_get_max_width(self) -> int:
        return _pxstr2int(self._ipywidget.layout.max_width)

    def _mgui_set_max_width(self, value: int | str):
        self._ipywidget.layout.max_width = _int2pxstr(value)

    def _mgui_get_height(self) -> int:
        """Return the current height of the widget."""
        return _pxstr2int(self._ipywidget.layout.height)

    def _mgui_set_height(self, value: int) -> None:
        """Set the current height of the widget."""
        self._ipywidget.layout.height = _int2pxstr(value)

    def _mgui_get_min_height(self) -> int:
        """Get the minimum allowable height of the widget."""
        return _pxstr2int(self._ipywidget.layout.min_height)

    def _mgui_set_min_height(self, value: int) -> None:
        """Set the minimum allowable height of the widget."""
        self._ipywidget.layout.min_height = _int2pxstr(value)

    def _mgui_get_max_height(self) -> int:
        """Get the maximum allowable height of the widget."""
        return _pxstr2int(self._ipywidget.layout.max_height)

    def _mgui_set_max_height(self, value: int) -> None:
        """Set the maximum allowable height of the widget."""
        self._ipywidget.layout.max_height = _int2pxstr(value)

    def _mgui_get_tooltip(self) -> str:
        return self._ipywidget.tooltip

    def _mgui_set_tooltip(self, value: str | None) -> None:
        self._ipywidget.tooltip = value

    def _ipython_display_(self, **kwargs):
        return self._ipywidget._ipython_display_(**kwargs)

    def _mgui_bind_parent_change_callback(self, callback):
        pass

    def _mgui_render(self):
        pass


class EmptyWidget(_IPyWidget):
    _ipywidget: ipywdg.Widget

    def _mgui_get_value(self) -> Any:
        raise NotImplementedError()

    def _mgui_set_value(self, value: Any) -> None:
        raise NotImplementedError()

    def _mgui_bind_change_callback(self, callback: Callable):
        pass


class _IPyValueWidget(_IPyWidget, protocols.ValueWidgetProtocol):
    def _mgui_get_value(self) -> float:
        return self._ipywidget.value

    def _mgui_set_value(self, value: Any) -> None:
        self._ipywidget.value = value

    def _mgui_bind_change_callback(self, callback):
        def _inner(change_dict):
            callback(change_dict.get("new"))

        self._ipywidget.observe(_inner, names=["value"])


class _IPyStringWidget(_IPyValueWidget):
    def _mgui_set_value(self, value) -> None:
        super()._mgui_set_value(str(value))


class _IPyRangedWidget(_IPyValueWidget, protocols.RangedWidgetProtocol):
    def _mgui_get_min(self) -> float:
        return self._ipywidget.min

    def _mgui_set_min(self, value: float) -> None:
        self._ipywidget.min = value

    def _mgui_get_max(self) -> float:
        return self._ipywidget.max

    def _mgui_set_max(self, value: float) -> None:
        self._ipywidget.max = value

    def _mgui_get_step(self) -> float:
        return self._ipywidget.step

    def _mgui_set_step(self, value: float) -> None:
        self._ipywidget.step = value

    def _mgui_get_adaptive_step(self) -> bool:
        return False

    def _mgui_set_adaptive_step(self, value: bool):
        # TODO:
        ...
        # raise NotImplementedError('adaptive step not implemented for ipywidgets')


class _IPySupportsOrientation(protocols.SupportsOrientation):
    _ipywidget: ipywdg.Widget

    def _mgui_set_orientation(self, value) -> None:
        self._ipywidget.orientation = value

    def _mgui_get_orientation(self) -> str:
        return self._ipywidget.orientation


class _IPySupportsChoices(protocols.SupportsChoices):
    _ipywidget: ipywdg.Widget

    def _mgui_get_choices(self) -> tuple[tuple[str, Any]]:
        """Get available choices."""
        return self._ipywidget.options

    def _mgui_set_choices(self, choices: Iterable[tuple[str, Any]]) -> None:
        """Set available choices."""
        options = [item for item in choices if item[1] is not Separator]
        self._ipywidget.options = options

    def _mgui_del_choice(self, choice_name: str) -> None:
        """Delete a choice."""
        options = [
            item
            for item in self._ipywidget.options
            if (not isinstance(item, tuple) or item[0] != choice_name)
            and item != choice_name
        ]
        self._ipywidget.options = options

    def _mgui_get_choice(self, choice_name: str) -> Any:
        """Get the data associated with a choice."""
        for item in self._ipywidget.options:
            if isinstance(item, tuple) and item[0] == choice_name:
                return item[1]
            elif item == choice_name:
                return item
        return None

    def _mgui_get_count(self) -> int:
        return len(self._ipywidget.options)

    def _mgui_get_current_choice(self) -> str:
        return self._ipywidget.label

    def _mgui_set_choice(self, choice_name: str, data: Any) -> None:
        """Set the data associated with a choice."""
        self._ipywidget.options = (*self._ipywidget.options, (choice_name, data))


class _IPySupportsText(protocols.SupportsText):
    """Widget that have text (in addition to value)... like buttons."""

    _ipywidget: ipywdg.Widget

    def _mgui_set_text(self, value: str) -> None:
        """Set text."""
        self._ipywidget.description = value

    def _mgui_get_text(self) -> str:
        """Get text."""
        return self._ipywidget.description


class _IPySupportsIcon(protocols.SupportsIcon):
    """Widget that can show an icon."""

    _ipywidget: ipywdg.Button

    def _mgui_set_icon(self, value: str | None, color: str | None) -> None:
        """Set icon."""
        # only ipywdg.Button actually supports icons.
        # but our button protocol allows it for all buttons subclasses
        # so we need this method in the concrete subclasses, but we
        # can't actually set the icon for anything but ipywdg.Button
        if hasattr(self._ipywidget, "icon"):
            # by splitting on ":" we allow for "prefix:icon-name" syntax
            # which works for iconify icons served by qt, while still
            # allowing for bare "icon-name" syntax which works for ipywidgets.
            # note however... only fa4/5 icons will work for ipywidgets.
            value = value or ""
            self._ipywidget.icon = value.replace("fa-", "").split(":", 1)[-1]
            self._ipywidget.style.text_color = color


class _IPyCategoricalWidget(_IPyValueWidget, _IPySupportsChoices):
    pass


class _IPyButtonWidget(_IPyValueWidget, _IPySupportsText, _IPySupportsIcon):
    pass


class _IPySliderWidget(_IPyRangedWidget, _IPySupportsOrientation):
    """Protocol for implementing a slider widget."""

    def __init__(self, readout: bool = True, orientation: str = "horizontal", **kwargs):
        super().__init__(**kwargs)

    def _mgui_set_readout_visibility(self, visible: bool) -> None:
        """Set visibility of readout widget."""
        # TODO

    def _mgui_get_tracking(self) -> bool:
        """If tracking is False, changed is only emitted when released."""
        # TODO
        return True

    def _mgui_set_tracking(self, tracking: bool) -> None:
        """If tracking is False, changed is only emitted when released."""
        # TODO


class Label(_IPyStringWidget):
    _ipywidget: ipywdg.Label


class LineEdit(_IPyStringWidget):
    _ipywidget: ipywdg.Text


class Password(_IPyStringWidget):
    _ipywidget: ipywdg.Password


class LiteralEvalLineEdit(_IPyStringWidget):
    _ipywidget: ipywdg.Text

    def _mgui_get_value(self) -> Any:
        from ast import literal_eval

        value = super()._mgui_get_value()
        return literal_eval(value)  # type: ignore


class TextEdit(_IPyStringWidget):
    _ipywidget: ipywdg.Textarea


class DateEdit(_IPyValueWidget):
    _ipywidget: ipywdg.DatePicker


class DateTimeEdit(_IPyValueWidget):
    _ipywidget: ipywdg.DatetimePicker


class TimeEdit(_IPyValueWidget):
    _ipywidget: ipywdg.TimePicker


class ToolBar(_IPyWidget):
    _ipywidget: ipywidgets.HBox

    def __init__(self, **kwargs):
        super().__init__(ipywidgets.HBox, **kwargs)
        self._icon_sz: tuple[int, int] | None = None

    def _mgui_add_button(self, text: str, icon: str, callback: Callable) -> None:
        """Add an action to the toolbar."""
        btn = ipywdg.Button(
            description=text, icon=icon, layout={"width": "auto", "height": "auto"}
        )
        if callback:
            btn.on_click(lambda e: callback())
        self._add_ipywidget(btn)

    def _add_ipywidget(self, widget: ipywidgets.Widget) -> None:
        children = list(self._ipywidget.children)
        children.append(widget)
        self._ipywidget.children = children

    def _mgui_add_separator(self) -> None:
        """Add a separator line to the toolbar."""
        # Define the vertical separator
        sep = ipywdg.Box(
            layout=ipywdg.Layout(border_left="1px dotted gray", margin="1px 4px")
        )
        self._add_ipywidget(sep)

    def _mgui_add_spacer(self) -> None:
        """Add a spacer to the toolbar."""
        self._add_ipywidget(ipywdg.Box(layout=ipywdg.Layout(flex="1")))

    def _mgui_add_widget(self, widget: Widget) -> None:
        """Add a widget to the toolbar."""
        self._add_ipywidget(widget.native)

    def _mgui_get_icon_size(self) -> tuple[int, int] | None:
        """Return the icon size of the toolbar."""
        return self._icon_sz

    def _mgui_set_icon_size(self, size: int | (tuple[int, int] | None)) -> None:
        """Set the icon size of the toolbar."""
        if isinstance(size, int):
            size = (size, size)
        elif size is None:
            size = (0, 0)
        elif not isinstance(size, tuple):
            raise ValueError("icon size must be an int or tuple of ints")
        sz = max(size)
        self._icon_sz = (sz, sz)
        for child in self._ipywidget.children:
            if hasattr(child, "style"):
                child.style.font_size = f"{sz}px" if sz else None
            child.layout.min_height = f"{sz*2}px" if sz else None

    def _mgui_clear(self) -> None:
        """Clear the toolbar."""
        self._ipywidget.children = ()


class PushButton(_IPyButtonWidget):
    _ipywidget: ipywdg.Button

    def _mgui_bind_change_callback(self, callback):
        self._ipywidget.on_click(lambda e: callback(False))

    # ipywdg.Button does not have any value. Return False for compatibility with Qt.
    def _mgui_get_value(self) -> float:
        return False

    def _mgui_set_value(self, value: Any) -> None:
        pass


class CheckBox(_IPyButtonWidget):
    _ipywidget: ipywdg.Checkbox


class RadioButton(_IPyButtonWidget):
    _ipywidget: ipywidgets.RadioButtons


class SpinBox(_IPyRangedWidget):
    _ipywidget: ipywidgets.IntText


class FloatSpinBox(_IPyRangedWidget):
    _ipywidget: ipywidgets.FloatText


class Slider(_IPySliderWidget):
    _ipywidget: ipywidgets.IntSlider


class FloatSlider(_IPySliderWidget):
    _ipywidget: ipywidgets.FloatSlider


class ComboBox(_IPyCategoricalWidget):
    _ipywidget: ipywidgets.Dropdown


class Select(_IPyCategoricalWidget):
    _ipywidget: ipywidgets.SelectMultiple


# CONTAINER ----------------------------------------------------------------------


class Container(_IPyWidget, protocols.ContainerProtocol, protocols.SupportsOrientation):
    def __init__(self, layout="horizontal", scrollable: bool = False, **kwargs):
        wdg_class = ipywidgets.VBox if layout == "vertical" else ipywidgets.HBox
        super().__init__(wdg_class, **kwargs)

    def _mgui_add_widget(self, widget: Widget) -> None:
        children = list(self._ipywidget.children)
        children.append(widget.native)
        self._ipywidget.children = children
        widget.parent = self._ipywidget

    def _mgui_insert_widget(self, position: int, widget: Widget) -> None:
        children = list(self._ipywidget.children)
        children.insert(position, widget.native)
        self._ipywidget.children = children
        widget.parent = self._ipywidget

    def _mgui_remove_widget(self, widget: Widget) -> None:
        children = list(self._ipywidget.children)
        children.remove(widget.native)
        self._ipywidget.children = children

    def _mgui_remove_index(self, position: int) -> None:
        children = list(self._ipywidget.children)
        children.pop(position)
        self._ipywidget.children = children

    def _mgui_count(self) -> int:
        return len(self._ipywidget.children)

    def _mgui_index(self, widget: Widget) -> int:
        return self._ipywidget.children.index(widget.native)

    def _mgui_get_index(self, index: int) -> Widget | None:
        """(return None instead of index error)."""
        return self._ipywidget.children[index]._magic_widget

    def _mgui_get_native_layout(self) -> Any:
        raise self._ipywidget

    def _mgui_get_margins(self) -> tuple[int, int, int, int]:
        margin = self._ipywidget.layout.margin
        if margin:
            try:
                top, rgt, bot, lft = (int(x.replace("px", "")) for x in margin.split())
                return lft, top, rgt, bot
            except ValueError:
                return margin
        return (0, 0, 0, 0)

    def _mgui_set_margins(self, margins: tuple[int, int, int, int]) -> None:
        lft, top, rgt, bot = margins
        self._ipywidget.layout.margin = f"{top}px {rgt}px {bot}px {lft}px"

    def _mgui_set_orientation(self, value) -> None:
        raise NotImplementedError(
            "Sorry, changing orientation after instantiation "
            "is not yet implemented for ipywidgets."
        )

    def _mgui_get_orientation(self) -> str:
        return "vertical" if isinstance(self._ipywidget, ipywdg.VBox) else "horizontal"


def get_text_width(text):
    # FIXME: how to do this in ipywidgets?
    return 40



import pytest
from psygnal import EmitLoopError

from magicgui import magic_factory
from magicgui.type_map._magicgui import MagicFactory
from magicgui.widgets import ComboBox, FunctionGui, Slider


def test_magic_factory():
    """Test basic magic_factory behavior."""

    @magic_factory(call_button=True)
    def factory(x: int = 1):
        return x

    assert isinstance(factory, MagicFactory)

    # factories make widgets that are FunctionGui instances
    widget1 = factory()
    assert isinstance(widget1, FunctionGui)
    assert widget1.call_button

    # You can call them repeatedly, and even override the initial kwargs
    # given to magic_factory (just like with functools.partial)
    widget2 = factory(call_button=False, x={"widget_type": "Slider"})
    assert widget1 is not widget2
    assert isinstance(widget2, FunctionGui)
    assert not widget2.call_button
    assert isinstance(widget2.x, Slider)

    # the widget, (like all FunctionGuis) is still callable and accepts args
    assert widget1() == 1
    assert widget1(3) == 3


def test_magic_factory_reuse():
    """Test magic_factory can be reused."""

    @magic_factory(x={"choices": ["a", "b"]})
    def factory(x="a"):
        return x

    # there was an earlier bug that overrode widget parameters.  this tests for that
    widget_a = factory()
    assert isinstance(widget_a.x, ComboBox)

    widget_b = factory()
    assert isinstance(widget_b.x, ComboBox)


def test_magic_factory_repr():
    """Test basic magic_factory behavior."""

    @magic_factory(labels=False)
    def factory(x: int = 1):
        return x

    rep = repr(factory)
    assert rep.startswith("MagicFactory(function=<function test_magic_factory_repr")
    assert rep.endswith("labels=False)")

    @magic_factory(labels=False, x={"widget_type": "Slider"})
    def factory2(x: int = 1):
        return x

    assert repr(factory2).endswith(
        "labels=False, param_options={'x': {'widget_type': 'Slider'}})"
    )


def test_magic_factory_only_takes_kwargs():
    @magic_factory(labels=False)
    def factory(x: int = 1):
        return x

    with pytest.raises(ValueError) as e:
        factory("positional")
        assert "only accept keyword arguments" in str(e)


@magic_factory
def self_referencing_factory(x: int = 1):
    """Function that refers to itself, and wants the FunctionGui instance."""
    return self_referencing_factory


def test_magic_factory_self_reference():
    """Test that self-referential factories work in global scopes."""
    widget = self_referencing_factory()
    assert isinstance(widget(), FunctionGui)


def test_magic_local_factory_self_reference():
    """Test that self-referential factories work in local scopes."""

    @magic_factory
    def local_self_referencing_factory(x: int = 1):
        return local_self_referencing_factory

    widget = local_self_referencing_factory()
    assert isinstance(widget(), FunctionGui)


def test_factory_init():
    def bomb():
        raise RuntimeError("boom")

    def widget_init(widget):
        widget.called.connect(bomb)

    @magic_factory(widget_init=widget_init)
    def factory(x: int = 1):
        pass

    widget = factory()

    with pytest.raises(EmitLoopError):
        widget()


def test_bad_value_factory_init():
    def widget_init():
        pass

    with pytest.raises(TypeError):

        @magic_factory(widget_init=widget_init)  # type: ignore
        def factory(x: int = 1):
            pass


def test_bad_type_factory_init():
    with pytest.raises(TypeError):

        @magic_factory(widget_init=1)  # type: ignore
        def factory(x: int = 1):
            pass


def test_none_defaults():
    """Make sure that an unannotated parameter with default=None is ok."""

    @magic_factory
    def factory(arg=None):
        return 1

    assert factory()() == 1



"""# Custom text labels for widgets

An example showing how to create custom text labels for your widgets.
"""

from magicgui import magicgui


# use a different label than the default (the parameter name) in the UI
@magicgui(x={"label": "widget to set x"})
def example(x=1, y="hi"):
    """Example function."""
    return x, y


example.changed.connect(print)
example.show(run=True)



"""https://www.mkdocs.org/dev-guide/plugins/#events ."""

from __future__ import annotations

import importlib.abc
import os
import sys
import types
import typing
import warnings
from contextlib import contextmanager
from importlib import import_module
from importlib.machinery import ModuleSpec
from itertools import count
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Mapping

from griffe.dataclasses import Alias
from griffe.docstrings import numpy
from mkdocstrings_handlers.python.handler import PythonHandler

from magicgui.type_map import get_widget_class

warnings.simplefilter("ignore", DeprecationWarning)

if TYPE_CHECKING:
    from mkdocs.structure.pages import Page


# TODO: figure out how to do this with options
@contextmanager
def _hide_numpy_warn():
    if not hasattr(numpy, "_warn"):
        yield
        return
    before, numpy._warn = numpy._warn, lambda *x, **k: None
    yield
    numpy._warn = before


def inject_dynamic_docstring(item: Alias, identifier: str) -> None:
    for i in range(1, 3):
        module_name, *names = identifier.rsplit(".", 1)
        try:
            module = import_module(module_name)
        except ModuleNotFoundError:
            continue
        else:
            obj = module
            for name in names:
                obj = getattr(obj, name)
            first_line, *rest = (obj.__doc__ or "").splitlines()
            if first_line and item.target.docstring:
                item.target.docstring.value = (
                    first_line + "\n" + dedent("\n".join(rest))
                )
            break


class WidgetHandler(PythonHandler):
    def collect(self, identifier: str, config: Mapping[str, Any]) -> Any:
        item = super().collect(identifier, config)
        if isinstance(item, Alias):
            inject_dynamic_docstring(item, identifier)
        # to edit default in the parameter table
        # item.parameters["something"].default = ...
        return item

    def render(self, data: Any, config: Mapping[str, Any]) -> str:
        with _hide_numpy_warn():
            return super().render(data, config)


class MyLoader(importlib.abc.Loader):
    def create_module(self, spec):
        return types.ModuleType(spec.name)

    def exec_module(self, module: types.ModuleType) -> None:
        def get_handler(
            *,
            theme: str,
            custom_templates: str | None = None,
            config_file_path: str | None = None,
            paths: list[str] | None = None,
            locale: str = "en",
            load_external_modules: bool | None = None,
            **config: Any,
        ) -> PythonHandler:
            return WidgetHandler(
                handler="python",
                theme=theme,
                custom_templates=custom_templates,
                config_file_path=config_file_path,
                paths=paths,
                locale=locale,
                load_external_modules=load_external_modules,
            )

        module.get_handler = get_handler


class Finder(importlib.abc.MetaPathFinder):
    def find_spec(self, fullname: str, *args, **kwargs) -> ModuleSpec | None:
        if fullname == "mkdocstrings_handlers.widget_handler":
            return ModuleSpec(fullname, MyLoader())
        return None


def on_startup(**kwargs):
    sys.meta_path.append(Finder())


def _replace_autosummary(md: str) -> str:
    lines = md.splitlines()
    start = lines.index("::: autosummary")
    try:
        last_line = lines.index("", start + 1)
    except ValueError:
        last_line = None
    table = ["| Widget | Description |", "| ---- | ----------- |"]
    for line in lines[start + 1 : last_line]:
        name = line.strip()
        if name:
            module, _name = name.rsplit(".", 1)
            obj = getattr(import_module(module), _name)
            table.append(f"| [`{_name}`][{name}] | {obj.__doc__.splitlines()[0]} |")
    lines[start:last_line] = table
    return "\n".join(lines)


def _replace_type_to_widget(md: str) -> str:
    lines = md.splitlines()
    start = lines.index("::: type_to_widget")
    try:
        last_line = lines.index("", start + 1)
    except ValueError:
        last_line = None
    table = [
        "| <div style='width:210px'>Type Hint</div> "
        "| <div style='width:100px'>Widget</div> "
        "| `__init__` kwargs",
        "| ---- | ------ | ------ |",
    ]
    for line in lines[start + 1 : last_line]:
        name = line.strip()
        if name:
            # eval Annotated types
            hint = eval(name, typing.__dict__) if "Annotated" in name else name
            wdg_type, kwargs = get_widget_class(annotation=hint)
            kwargs.pop("nullable", None)
            kwargs = f"`{kwargs}`" if kwargs else ""
            wdg_name = f"magicgui.widgets.{wdg_type.__name__}"
            wdg_link = f"[`{wdg_type.__name__}`][{wdg_name}]"
            if "[" in name:
                _name = name.split("[")[0]
                name_link = f"[`{name}`][typing.{_name}]"
            else:
                name_link = f"[`{name}`][{name}]"
            table.append(f"| {name_link}  | {wdg_link} | {kwargs} | ")

    lines[start:last_line] = table
    return "\n".join(lines)


def _replace_widget_autosummary(md: str) -> str:
    from magicgui import widgets

    lines = md.splitlines()
    start = lines.index("::: widget_autosummary")
    try:
        last_line = lines.index("", start + 1)
    except ValueError:
        last_line = None

    autosummary = ["::: autosummary"]
    for name in dir(widgets):
        if name.startswith("_"):
            continue
        obj = getattr(widgets, name)
        if (
            isinstance(obj, type)
            and issubclass(obj, widgets.Widget)
            and obj is not widgets.Widget
            and obj.__name__ == name
        ):
            autosummary.append(f"magicgui.widgets.{name}")

    lines[start:last_line] = autosummary
    return "\n".join(lines)


def on_page_markdown(md: str, page: Page, **kwargs: Any) -> str:
    """Called when mkdocs is building the markdown for a page."""
    import re

    w_iter = count()

    while "::: widget_autosummary" in md:
        md = _replace_widget_autosummary(md)

    while "::: autosummary" in md:
        md = _replace_autosummary(md)

    while "::: type_to_widget" in md:
        md = _replace_type_to_widget(md)

    def _add_images(matchobj: re.Match) -> str:
        # add image links below all code blocks with a `.show()` call
        if ".show()" not in matchobj.group(0):
            return matchobj.group(0) or ""

        src = matchobj.group(1)  # source code
        reldepth = "../" * page.file.src_path.count(os.sep)  # relative of this file
        dest = f"{reldepth}_images/{page.file.name}_{next(w_iter)}.png"  # link to img
        link = f"\n![]({dest}){{: .code-image}}\n\n"  # theme aware link

        # Not working ... theme aware
        # light = dest.replace(".png", "_light.png")
        # dark = dest.replace(".png", "_light.png")
        # new_md += f"\n![]({dark}#only-dark){{: .code-image}}\n\n"
        new_md: str = "```python\n" + src + "\n```" + link
        return new_md.replace(" # leave open", "")

    if page.title not in {"migration guide"}:
        md = re.sub("``` ?python\n([^`]*)```", _add_images, md, re.DOTALL)

    return md



#!/usr/bin/env python

"""Tests for `magicgui._qt` module."""


import pytest
from qtpy import API_NAME, QtCore
from qtpy import QtWidgets as QtW

from magicgui import event_loop, widgets


@pytest.mark.skipif("PyQt" in API_NAME, reason="Couldn't delete app on PyQt")
def test_event():
    """Test that the event loop makes a Qt app."""
    if QtW.QApplication.instance():
        if API_NAME == "PySide2":
            __import__("shiboken2").delete(QtW.QApplication.instance())
        elif API_NAME == "PySide6":
            __import__("shiboken6").delete(QtW.QApplication.instance())
        else:
            raise AssertionError(f"known API name: {API_NAME}")
    assert not QtW.QApplication.instance()
    with event_loop():
        app = QtW.QApplication.instance()
        assert app
        timer = QtCore.QTimer()
        timer.timeout.connect(lambda: app.exit())
        timer.start(100)


@pytest.mark.parametrize(
    "WidgetClass",
    [
        widgets.ComboBox,
        widgets.LineEdit,
        widgets.PushButton,
        widgets.DateTimeEdit,
        widgets.SpinBox,
        widgets.FloatSpinBox,
        widgets.Slider,
        widgets.FloatSlider,
        widgets.FileEdit,
    ],
)
def test_get_set_change(WidgetClass):
    """Test that we can retrieve getters, setters, and signals for most Widgets."""
    _ = WidgetClass()



"""# File dialog widget

A dialog to select multiple files.
"""

from pathlib import Path
from typing import Sequence

from magicgui import magicgui


# Sequence of paths
# We change the label using "label" for added clarity
# the filter argument restricts file types
@magicgui(filenames={"label": "Choose Tiff files:", "filter": "*.tif"})
def filespicker(filenames: Sequence[Path]) -> Sequence[Path]:
    """Take a filename and do something with it."""
    print("The filenames are:", filenames)
    return filenames


filespicker.filenames.changed.connect(print)
filespicker.show(run=True)



"""# Deocrate class methods with magicgui

Demonstrates decorating a class method with magicgui.

Once the class is instantiated, `instance.method_name` will return a FunctionGui
in which the instance will always be provided as the first argument (i.e. "self") when
the FunctionGui or method is called.
"""

from magicgui import event_loop, magicgui
from magicgui.widgets import Container


class MyObject:
    """Example object class."""

    def __init__(self, name):
        self.name = name
        self.counter = 0.0

    @magicgui(auto_call=True)
    def method(self, sigma: float = 0):
        """Example class method."""
        print(f"instance: {self.name}, counter: {self.counter}, sigma: {sigma}")
        self.counter = self.counter + sigma
        return self.name


with event_loop():
    a = MyObject("a")
    b = MyObject("b")
    container = Container(widgets=[a.method, b.method])
    container.show()
    assert a.method() == "a"
    assert b.method() == "b"
